/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <errno.h>

#include "globals.h"
#include "main.h"
#include "ftp.h"
#include "fxp.h"
#include "queue.h"

/* Local functions */
static void fxp_event(int signo);
gint fxp_event_handler(gpointer data);

static GList *fxps = NULL;

void init_fxp()
{
  struct sigaction act;

  /* setup signal handler for SIGUSR1 which is used to tell the
   * other deadftp proc that new data is avail
   */
  act.sa_handler = fxp_event;
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_RESTART;
  if (sigaction(SIGUSR1, &act, NULL) < 0) {
    g_warning("Couldn't setup signal handler\n");
    return;
  }
}

fxp_desc_t *fxp_get()
{
  fxp_desc_t *fxp = g_new0(fxp_desc_t, 1);

  fxp->shm_size = FXP_SHM_SIZE;
  fxp->shm_id = shmget(IPC_PRIVATE, fxp->shm_size, IPC_CREAT | 0700);
  fxp->slave = FALSE;
  fxp->refcount = 1;
  if (fxp->shm_id == -1) {
    g_error("Couldn't allocate shared memory: %s\n", g_strerror(errno));
    g_free(fxp);
    return NULL;
  }

  fxp->shm_area = shmat(fxp->shm_id, NULL, 0);
  if (fxp->shm_area == (unsigned char *) -1) {
    g_error("Couldn't attach to shared memory: %s\n", g_strerror(errno));
    g_free(fxp);
    return NULL;
  }

  fxp->data = (fxp_t *) fxp->shm_area;
  /* a lock is used to ensure only one process is accessing the shared
   * data at a time */
  pthread_mutex_init(&fxp->data->lock, NULL);
  /* no other process can access this data yet so we don't
   * need to lock the lock when setting this */
  fxp->data->master_pid = getpid();
  FXP_STATUS(fxp) = FXP_STATUS_IDLE;

  fxps = g_list_append(fxps, fxp);

  return fxp;
}

fxp_desc_t *fxp_attach(int shm_id)
{
  fxp_desc_t *fxp;
  GList *l;

  fxp = g_new0(fxp_desc_t, 1);

  fxp->shm_id = shm_id;
  fxp->slave = TRUE;
  fxp->shm_area = shmat(fxp->shm_id, NULL, 0);
  fxp->refcount = 1;
  if (fxp->shm_area == (unsigned char *) -1) {
    g_error("Couldn't attach to shared memory: %s\n", g_strerror(errno));
    g_free(fxp);
    return NULL;
  }

  fxp->data = (fxp_t *) fxp->shm_area;
  /* The mutex should have been initialized by the master process */
  pthread_mutex_lock(&fxp->data->lock);
  fxp->data->slave_pid = getpid();
  pthread_mutex_unlock(&fxp->data->lock);

  /* try to find an existing fxp_desc with the master process to
   * conserve shm */
  pthread_mutex_lock(&fxp->data->lock);
  for (l = fxps; l; l = g_list_next(l)) {
    fxp_desc_t *f = (fxp_desc_t *) l->data;
    pthread_mutex_lock(&f->data->lock);
    if (f->data->master_pid == fxp->data->master_pid) {
      /* set value to the shm_id of the other segment */
      fxp->data->value = f->shm_id;
      pthread_mutex_unlock(&f->data->lock);
      pthread_mutex_unlock(&fxp->data->lock);
      /* tell the master process to use the other shm */
      fxp_set_status(fxp, FXP_STATUS_USE_OTHER);

      shmdt(fxp->shm_area);
      g_free(fxp);

      f->refcount++;
      return f;
    }
    pthread_mutex_unlock(&f->data->lock);
  }
  pthread_mutex_unlock(&fxp->data->lock);

  fxps = g_list_append(fxps, fxp);

  return fxp;
}

void fxp_event(int signo)
{
  /* It's unsafe to do the work from inside the signal handler (causes
   * async errors from XLib) so set a timer to handle the signal after 
   * the handler returns.  */

  gtk_timeout_add(1, fxp_event_handler, NULL);
}

gint fxp_event_handler(gpointer data)
{
  GList *list = fxps;

  while (list) {
    fxp_desc_t *fxp = (fxp_desc_t *) list->data;
    unsigned long status;

    if (!FXP_DIRTY(fxp)) {
      list = g_list_next(list);
      continue;
    }

    pthread_mutex_lock(&fxp->data->lock);
    FXP_CLEAN(fxp);
    status = FXP_STATUS(fxp);
    pthread_mutex_unlock(&fxp->data->lock);

    switch (status) {
    case FXP_STATUS_USE_OTHER:
      {
	GList *find;

	for (find = fxps; find; find = g_list_next(find)) {
	  fxp_desc_t *f = (fxp_desc_t *) find->data;
	  /* value is the shmid of the other segment to use */
	  if (fxp->data->value == f->shm_id) {
	    GList *queue;

	    /* change the fxp on all the queue items to the other fxp item */
	    for (queue = ftp_queue; queue; queue = g_list_next(queue)) {
	      queue_item_t *q = (queue_item_t *) queue;
	      if (q->fxp == fxp) {
		f->refcount++;
		q->fxp = f;
	      }
	    }
	    fxp->refcount--;
	    if (!fxp->refcount) {
	      shmdt(fxp->shm_area);
	      g_free(fxp);
	    }
	  }
	}
	break;
      }
    case FXP_STATUS_START:
      if (FTP.state != FTP_STATE_IDLE)
	fxp_set_status(fxp, FXP_STATUS_BUSY);
      else if (fxp->slave)
	fxp_set_status(fxp, FXP_STATUS_OK);
      else {
	queueid_t qid;
	pthread_mutex_lock(&fxp->data->lock);
	/* queueid is stored in value */
	qid = fxp->data->value;
	pthread_mutex_unlock(&fxp->data->lock);

	queue_go_id(FALSE, qid);
      }
      break;
    case FXP_STATUS_BUSY:
      /* move on to the next queue item */
      QUEUE_SET_STATUS(QUEUE_CURRENT(FTP.queue), QUEUE_STATUS_PAUSE);
      queue_start();
      break;
    case FXP_STATUS_OK:
      /* slave response to master when master requested fxp start */
      queue_start();
      break;
    case FXP_STATUS_PORT:
      /* master sending port info to slave.  Make sure we're connected to
       * the right host and in the right dir by calling queue_go */
      queue_start();
      break;
    case FXP_STATUS_GO:
      /* sent from the slave to the master to indicate connection was
       * setup up successfully.  Master now needs to start the transfer. */
      g_assert(QUEUE_CURRENT(FTP.queue)->fxp == fxp);
      ftp_send_command(FTP_STATE_FXP_TRANS, 0, _("Tranferring File"), "RETR %s",
		       QUEUE_CURRENT(FTP.queue)->remote_file);
      fxp_set_status(fxp, FXP_STATUS_TRANS);
    case FXP_STATUS_TRANS:
      break;
    case FXP_STATUS_STOP:
      ftp_send_command(FTP_STATE_ABORT, 0, _("Aborting"), "ABOR");
      break;
    case FXP_STATUS_DELETE:
      queue_item_delete_id(fxp->slave, fxp->data->value);
      break;
    default:
      break;
    }

    list = g_list_next(list);
  }
  return FALSE;
}

gboolean fxp_start_transfer(queue_item_t * q)
{
  pthread_mutex_lock(&q->fxp->data->lock);
  /* If the status isn't IDLE then don't try to start a new connection */
  if (FXP_STATUS(q->fxp) != FXP_STATUS_IDLE) {
    pthread_mutex_unlock(&q->fxp->data->lock);
    return FALSE;
  }
  /* value is used to hold the queueid */
  q->fxp->data->value = q->queueid;
  pthread_mutex_unlock(&q->fxp->data->lock);

  fxp_set_status(q->fxp, FXP_STATUS_START);
  return TRUE;
}

void fxp_set_status(fxp_desc_t * fxp, unsigned long status)
{
  pthread_mutex_lock(&fxp->data->lock);
  /* mutex must be locked before calling this function! */
  FXP_STATUS(fxp) = status;
  FXP_SET_DIRTY(fxp);
  if (fxp->slave)
    kill(fxp->data->master_pid, SIGUSR1);
  else
    kill(fxp->data->slave_pid, SIGUSR1);
  pthread_mutex_unlock(&fxp->data->lock);
}

void fxp_delete(queue_item_t * q)
{
  q->fxp->refcount--;

  if (q->fxp->refcount == 0) {
    q->fxp->data->value = q->queueid;
    fxp_set_status(q->fxp, FXP_STATUS_DELETE);

    shmdt(q->fxp->shm_area);
    fxps = g_list_remove(fxps, q->fxp);
    g_free(q->fxp);

    g_warning("remove fxp_desc\n");
  }
}
