/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>

#include "globals.h"
#include "main.h"
#include "hostmanager.h"
#include "ftp.h"
#include "support.h"

/* Callbacks in this file */
void on_add_host_clicked(GtkButton * button, gpointer user_data);
void on_add_group_clicked(GtkButton * button, gpointer user_data);
void on_group_ok_clicked(GtkButton * button, gpointer user_data);
void on_group_cancel_clicked(GtkButton * button, gpointer user_data);
void on_hm_connect_clicked(GtkButton * button, gpointer user_data);
void on_hm_connect_clicked(GtkButton * button, gpointer user_data);
void on_hm_exit_clicked(GtkButton * button, gpointer user_data);
void on_hm_apply_clicked(GtkButton * button, gpointer user_data);
void on_anonymous_toggled(GtkToggleButton * togglebutton, gpointer user_data);
void on_delete_host_clicked(GtkButton * button, gpointer user_data);
void on_delete_group_clicked(GtkButton * button, gpointer user_data);
void on_host_tree_select_child(GtkTree * tree, GtkWidget * widget,
			       gpointer user_data);
void on_hm_new_clicked(GtkButton * button, gpointer user_data);
void on_sitename_ok_clicked(GtkButton * button, gpointer user_data);
void on_sitename_cancel_clicked(GtkButton * button, gpointer user_data);
void on_hostmanager_destroy(GtkObject * object, gpointer user_data);
gboolean on_host_ctree_button_press_event(GtkWidget * widget, GdkEventButton * event,
					  gpointer user_data);
void on_host_ctree_tree_select_row(GtkCTree * ctree, GList * node, gint column,
				   gpointer user_data);
void on_add_cur_host_clicked(GtkButton * button, gpointer user_data);
void on_hostmanag_clicked(GtkButton * button, gpointer user_data);
void on_hostmanager1_activate(GtkMenuItem * menuitem, gpointer user_data);
void hostmanager_changed(void);

/* Local Functions */
static void hm_check_and_update(xmlNodePtr host, const char *attr,
				const char *newvalue);
static xmlNodePtr find_Host_Node(xmlNodePtr host, const char *n);
static void hm_add_hostgroup_to_tree(GtkWidget * tree, xmlNodePtr node);
static void hm_tree_add_group(GtkWidget *, xmlNodePtr);
static void hm_tree_add_host(GtkWidget * group, xmlNodePtr node);
static void hm_tree_delete_group(GtkWidget * tree, GtkWidget * grp);
static void hm_tree_delete_host(GtkWidget * item, GtkWidget * subitem);
static void hm_ctree_add_host(GtkWidget * tree, GtkCTreeNode * parent,
			      xmlNodePtr node);
static void hm_ctree_add_group(GtkWidget * tree, GtkCTreeNode * parent,
			       xmlNodePtr node);
void start_hostmanager(void);
void hm_redisplay_tree(void);
void hm_add_group(const char *group);
void hm_add_host(const char *group, const char *sitename, const char *hostname,
		 const char *port, const char *username, int anonymous,
		 const char *password, const char *local, const char *remote,
		 const char *notes, int timeout, int retry_delay, int retry_count,
		 int show_hidden, int use_pasv, int keepalive);
void delete_hostmanager(void);
void hostmanager_is_gone(void);
void hm_delete_host(GtkWidget * tree, GtkWidget * item);
void hm_delete_group(GtkWidget * item);
void hm_change_host(GtkWidget * item);
void hm_clear_host(void);
void hm_new_site(const char *name);
void hm_change_host_info(GtkWidget * item);
void check_and_update_toggle(xmlNodePtr, const char *attr, int newvalue);
void check_and_update_char_toggle(GtkWidget * item, const char *attr,
				  const char *newvalue);
int hm_is_group(GtkWidget * item);
void hm_ctree_connect(GtkCTreeNode * selected_node);

extern void change_local(void);

static xmlDocPtr doc;
static xmlNsPtr ns;
static gboolean ctree_need_redraw = FALSE;

GtkCTreeNode *ctree_selected_node;

hostmanager_t hostmanager;

void init_hostmanager()
{
  FILE *f;
  char filename[SMALL_BUFFER], *home;

  home = g_get_home_dir();
  if (!home) {
    return;
  }
  g_snprintf(filename, SMALL_BUFFER, "%s/.dead", home);
  mkdir(filename, S_IRUSR | S_IWUSR | S_IXUSR);
  g_snprintf(filename, SMALL_BUFFER, "%s/.dead/hostmanager", home);

  f = fopen(filename, "r");
  if (f != NULL) {
    int res, size = REGULAR_BUFFER;
    char chars[REGULAR_BUFFER];
    xmlParserCtxtPtr ctxt;

    /* Opened file, read and parse the XML document */
    res = fread(chars, 1, 4, f);
    if (res > 0) {
      ctxt = xmlCreatePushParserCtxt(NULL, NULL, chars, res, filename);
      while ((res = fread(chars, 1, size, f)) > 0) {
	xmlParseChunk(ctxt, chars, res, 0);
      }
      xmlParseChunk(ctxt, chars, 0, 1);
      doc = ctxt->myDoc;
      xmlFreeParserCtxt(ctxt);
    }
    /* close the file */
    fclose(f);
    if (!doc->root || !doc->root->name
	|| g_strcasecmp(doc->root->name, "HostManager") != 0) {
      xmlNodePtr tree;
      /* Not a valid XML document so create a new DOM structure */
      xmlFreeDoc(doc);
      doc = xmlNewDoc("1.0");
      doc->root = xmlNewDocNode(doc, ns, "HostManager", NULL);
      xmlSetNs(doc->root, ns);
      tree = xmlNewChild(doc->root, ns, "HostGroup", NULL);
      xmlSetProp(tree, "name", "General");
    } else {
      /* Look for the correct namespace */
      ns =
	xmlSearchNsByHref(doc, doc->root,
			  "http://deadftp.sourceforge.net/hostmanager");
      if (!ns) {
	g_print(_("Hostmanager file has wrong namespace\n"));
	ns = xmlNewNs(doc->root, "http://deadftp.sourceforge.net/hostmanager", "hm");
	xmlSetNs(doc->root, ns);
      }
    }
  } else {
    xmlNodePtr tree;
    doc = xmlNewDoc("1.0");
    doc->root = xmlNewDocNode(doc, NULL, "HostManager", NULL);
    ns = xmlNewNs(doc->root, "http://deadftp.sourceforge.net/hostmanager", "hm");
    xmlSetNs(doc->root, ns);
    tree = xmlNewChild(doc->root, ns, "HostGroup", NULL);
    xmlSetProp(tree, "name", "General");
  }
  hostmanager.window = NULL;
}
static void add_hosts_to_tree(GtkWidget * item, xmlNodePtr node)
{
  GtkWidget *subtree, *t, *box, *pixmap, *label;
  xmlNodePtr n;
  char *ch;

  subtree = gtk_tree_new();
  gtk_signal_connect(GTK_OBJECT(subtree), "select_child",
		     GTK_SIGNAL_FUNC(on_host_tree_select_child), NULL);

  gtk_tree_set_selection_mode(GTK_TREE(subtree), GTK_SELECTION_SINGLE);
  gtk_tree_item_set_subtree(GTK_TREE_ITEM(item), subtree);
  n = node->childs;
  while (n) {
    if (g_strcasecmp(n->name, "HostGroup") == 0) {
      hm_add_hostgroup_to_tree(subtree, n);
      n = n->next;
      continue;
    } else if (g_strcasecmp(n->name, "Host") != 0) {
      n = n->next;
      continue;
    }
    t = gtk_tree_item_new();
    box = gtk_hbox_new(FALSE, 0);
    pixmap = load_pixmap(hostmanager.window, "host.xpm", 0);
    gtk_box_pack_start(GTK_BOX(box), pixmap, FALSE, FALSE, 0);
    gtk_widget_show(pixmap);
    ch = xmlGetProp(n, "name");
    label = gtk_label_new(ch);
    g_free(ch);
    gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, 0);
    gtk_widget_show(label);
    gtk_container_add(GTK_CONTAINER(t), box);
    gtk_widget_show(box);

    gtk_object_set_data(GTK_OBJECT(t), "Label", (gpointer) xmlGetProp(n, "name"));
    gtk_object_set_data(GTK_OBJECT(t), "HostGroup", (gpointer) item);
    gtk_object_set_data(GTK_OBJECT(t), "Node", n);
    gtk_tree_append(GTK_TREE(subtree), t);
    gtk_widget_show(t);
    n = n->next;
  }
}
void start_hostmanager()
{
  GladeXML *xml;

  if (hostmanager.window)
    return;
  xml = glade_xml_new(deadftp.gladefile, "hostmanager");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  hostmanager.window = glade_xml_get_widget(xml, "hostmanager");

  STORE_WIDGET(hostmanager, host_tree);

  STORE_WIDGET(hostmanager, add_host);
  STORE_WIDGET(hostmanager, delete_host);
  STORE_WIDGET(hostmanager, add_group);
  STORE_WIDGET(hostmanager, delete_group);

  STORE_WIDGET(hostmanager, sitename);
  STORE_WIDGET(hostmanager, hostname);
  STORE_WIDGET(hostmanager, port);
  STORE_WIDGET(hostmanager, username);
  STORE_WIDGET(hostmanager, anonymous);
  STORE_WIDGET(hostmanager, password);
  STORE_WIDGET(hostmanager, local_path);
  STORE_WIDGET(hostmanager, remote_path);
  STORE_WIDGET(hostmanager, notes);

  STORE_WIDGET(hostmanager, timeout);
  STORE_WIDGET(hostmanager, retry_count);
  STORE_WIDGET(hostmanager, retry_delay);

  STORE_WIDGET(hostmanager, show_hidden);
  STORE_WIDGET(hostmanager, use_pasv);
  STORE_WIDGET(hostmanager, keepalive);

  STORE_WIDGET(hostmanager, resume_always);
  STORE_WIDGET(hostmanager, resume_ask);
  STORE_WIDGET(hostmanager, resume_never);

  STORE_WIDGET(hostmanager, hm_apply);

  hostmanager.changed = FALSE;
  gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.hm_apply), FALSE);

  glade_xml_signal_autoconnect(xml);

  gtk_object_unref(GTK_OBJECT(xml));

  gtk_object_set_data(GTK_OBJECT(hostmanager.host_tree), "Node",
		      (gpointer) doc->root);

  /* Anonymous is set by default */
  gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.username), FALSE);
  gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.password), FALSE);

  hm_redisplay_tree();
}
static void hm_add_hostgroup_to_tree(GtkWidget * tree, xmlNodePtr node)
{
  GtkWidget *t, *box, *pixmap, *label;
  char *ch;

  t = gtk_tree_item_new();
  box = gtk_hbox_new(FALSE, 0);
  pixmap = load_pixmap(hostmanager.window, "hostgroup.xpm", 0);
  gtk_box_pack_start(GTK_BOX(box), pixmap, FALSE, FALSE, 0);
  gtk_widget_show(pixmap);
  ch = xmlGetProp(node, "name");
  label = gtk_label_new(ch);
  g_free(ch);
  gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  gtk_container_add(GTK_CONTAINER(t), box);
  gtk_widget_show(box);

  ch = xmlGetProp(node, "name");
  gtk_object_set_data(GTK_OBJECT(tree), ch, (gpointer) t);
  gtk_object_set_data(GTK_OBJECT(t), "Label", ch);
  g_free(ch);
  gtk_object_set_data(GTK_OBJECT(t), "Node", node);
  gtk_tree_append(GTK_TREE(tree), t);
  gtk_widget_show(t);
  add_hosts_to_tree(t, node);
}

void hm_redisplay_tree()
{
  xmlNodePtr node;

  g_return_if_fail(hostmanager.window != NULL);

  gtk_tree_clear_items(GTK_TREE(hostmanager.host_tree), 0,
		       g_list_length(GTK_TREE(hostmanager.host_tree)->children));

  node = doc->root->childs;
  while (node) {
    if (g_strcasecmp(node->name, "HostGroup") != 0) {
      node = node->next;
      continue;
    }
    hm_add_hostgroup_to_tree(hostmanager.host_tree, node);
    node = node->next;
  }
}
void save_hostmanager()
{
  char filename[SMALL_BUFFER], *home;

  home = g_get_home_dir();
  if (!home) {
    return;
  }
  g_snprintf(filename, SMALL_BUFFER, "%s/.dead/hostmanager", home);

  xmlSaveFile(filename, doc);
  chmod(filename, S_IRUSR | S_IWUSR);
}
void hm_add_group(const char *group)
{
  GtkWidget *item;
  xmlNodePtr grp;
  GList *sel;
  xmlNodePtr node;

  g_return_if_fail(hostmanager.window != NULL);

  sel = GTK_TREE_SELECTION(GTK_TREE(hostmanager.host_tree));
  if (!sel) {
    item = NULL;
    node = doc->root;
  } else {
    item = GTK_WIDGET(sel->data);
    if (gtk_object_get_data(GTK_OBJECT(item), "HostGroup"))
      return;
    node = (xmlNodePtr) gtk_object_get_data(GTK_OBJECT(item), "Node");
  }
  g_return_if_fail(node != NULL);

  grp = xmlNewChild(node, ns, "HostGroup", NULL);
  xmlSetProp(grp, "name", group);

  ctree_need_redraw = TRUE;
  if (item) {
    hm_tree_add_group(item, grp);
  }
}
void hm_add_host(const char *group, const char *sitename, const char *hostname,
		 const char *port, const char *username, int anonymous,
		 const char *password, const char *local, const char *remote,
		 const char *notes, int timeout, int retry_delay, int retry_count,
		 int show_hidden, int use_pasv, int keepalive)
{
  xmlNodePtr grp, host, sub1;
  GtkWidget *t;
  GList *sel;

  if (!sitename || !*sitename)
    return;

  g_return_if_fail(hostname && *hostname);

  if (hostmanager.window) {
    sel = GTK_TREE_SELECTION(GTK_TREE(hostmanager.host_tree));

    if (!sel)
      return;

    t = GTK_WIDGET(sel->data);
    if (!t)
      return;
    if (gtk_object_get_data(GTK_OBJECT(t), "HostGroup"))	/* It's a Host entry */
      return;
    grp = gtk_object_get_data(GTK_OBJECT(t), "Node");
  } else {
    grp = find_Host_Node(doc->root, "HostGroup");	/* Add it to the first Group */
    t = NULL;
  }

  g_return_if_fail(grp != NULL);

  host = xmlNewChild(grp, ns, "Host", NULL);
  xmlSetProp(host, "name", sitename);
  xmlNewTextChild(host, ns, "hostname", hostname);

  sub1 = xmlNewChild(host, ns, "port", NULL);
  xmlSetProp(sub1, "value", port);

  if (anonymous) {
    sub1 = xmlNewChild(host, ns, "anonymous", NULL);
    xmlSetProp(sub1, "value", "1");
  } else {
    sub1 = xmlNewChild(host, ns, "anonymous", NULL);
    xmlSetProp(sub1, "value", "0");

    xmlNewTextChild(host, ns, "username", username);
    xmlNewTextChild(host, ns, "password", password);
  }
  xmlNewTextChild(host, ns, "local_dir", local);
  xmlNewTextChild(host, ns, "remote_dir", remote);
  xmlNewTextChild(host, ns, "notes", notes);

  ctree_need_redraw = TRUE;

  if (t) {
    hm_tree_add_host(t, host);
  }
}

void delete_hostmanager()
{
  gtk_widget_destroy(hostmanager.window);
}

void hostmanager_is_gone()
{
  hostmanager.window = NULL;
}

void hm_delete_host(GtkWidget * tree, GtkWidget * item)
{
  xmlNodePtr node;

  node = gtk_object_get_data(GTK_OBJECT(item), "Node");

  g_return_if_fail(node != NULL);

  xmlUnlinkNode(node);
  xmlFreeNode(node);

  hm_tree_delete_host(tree, item);
}

void hm_delete_group(GtkWidget * item)
{
  xmlNodePtr node;
  GtkWidget *tree = item->parent;

  node = gtk_object_get_data(GTK_OBJECT(item), "Node");

  g_return_if_fail(node != NULL);

  xmlUnlinkNode(node);
  xmlFreeNode(node);

  hm_tree_delete_group(tree, item);
}

static void hm_tree_delete_host(GtkWidget * item, GtkWidget * subitem)
{
  GList *l;
  GtkWidget *subtree;

  g_return_if_fail(item != NULL);
  g_return_if_fail(subitem != NULL);

  subtree = GTK_TREE_ITEM_SUBTREE(item);
  g_return_if_fail(subtree != NULL);

  l = g_list_append(NULL, (gpointer) subitem);
  gtk_tree_remove_items(GTK_TREE(subtree), l);
}
static void hm_tree_add_host(GtkWidget * group, xmlNodePtr node)
{
  GtkWidget *t, *box, *pixmap, *label;
  char *name;

  g_return_if_fail(node != NULL);
  g_return_if_fail(group != NULL);

  t = gtk_tree_item_new();
  box = gtk_hbox_new(FALSE, 0);
  pixmap = load_pixmap(hostmanager.window, "host.xpm", 0);
  gtk_box_pack_start(GTK_BOX(box), pixmap, FALSE, FALSE, 0);
  gtk_widget_show(pixmap);
  name = xmlGetProp(node, "name");
  label = gtk_label_new(name);
  gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  gtk_container_add(GTK_CONTAINER(t), box);
  gtk_widget_show(box);
  gtk_widget_show(t);

  if (!GTK_TREE_ITEM_SUBTREE(group)) {
    GtkWidget *subtree = gtk_tree_new();
    gtk_signal_connect(GTK_OBJECT(subtree), "select_child",
		       GTK_SIGNAL_FUNC(on_host_tree_select_child), NULL);
    gtk_tree_set_selection_mode(GTK_TREE(subtree), GTK_SELECTION_SINGLE);
    gtk_tree_item_set_subtree(GTK_TREE_ITEM(group), subtree);
  }
  /* name is set above */
  gtk_object_set_data(GTK_OBJECT(t), "Label", name);
  g_free(name);
  gtk_object_set_data(GTK_OBJECT(t), "HostGroup", (gpointer) group);
  gtk_object_set_data(GTK_OBJECT(t), "Node", node);
  gtk_tree_append(GTK_TREE(GTK_TREE_ITEM_SUBTREE(group)), t);
}
static void hm_tree_add_group(GtkWidget * tree, xmlNodePtr node)
{
  GtkWidget *t, *box, *pixmap, *label;
  char *name;

  g_return_if_fail(tree != NULL);
  g_return_if_fail(node != NULL);

  t = gtk_tree_item_new();
  box = gtk_hbox_new(FALSE, 0);
  pixmap = load_pixmap(hostmanager.window, "hostgroup.xpm", 0);
  gtk_widget_show(pixmap);
  gtk_box_pack_start(GTK_BOX(box), pixmap, FALSE, FALSE, 0);
  name = xmlGetProp(node, "name");
  label = gtk_label_new(name);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, 0);
  gtk_container_add(GTK_CONTAINER(t), box);
  gtk_widget_show(box);

  /* name set above */
  gtk_object_set_data(GTK_OBJECT(t), "Label", name);
  gtk_object_set_data(GTK_OBJECT(t), "Node", node);
  if (tree) {
    if (!GTK_TREE_ITEM_SUBTREE(tree)) {
      GtkWidget *subtree = gtk_tree_new();
      gtk_signal_connect(GTK_OBJECT(subtree), "select_child",
			 GTK_SIGNAL_FUNC(on_host_tree_select_child), NULL);
      gtk_tree_set_selection_mode(GTK_TREE(subtree), GTK_SELECTION_SINGLE);
      gtk_tree_item_set_subtree(GTK_TREE_ITEM(tree), subtree);
    }
    /* name set above */
    gtk_object_set_data(GTK_OBJECT(tree), name, (gpointer) t);
    gtk_tree_append(GTK_TREE(GTK_TREE_ITEM_SUBTREE(tree)), t);
  } else {
    gtk_object_set_data(GTK_OBJECT(hostmanager.host_tree), name, (gpointer) t);
    gtk_tree_append(GTK_TREE(hostmanager.host_tree), t);
  }
  g_free(name);
  gtk_widget_show(t);
}
static void hm_tree_delete_group(GtkWidget * tree, GtkWidget * grp)
{
  GList *l = g_list_append(NULL, (gpointer) grp);
  gpointer label;

  g_return_if_fail(tree != NULL);
  g_return_if_fail(grp != NULL);

  label = gtk_object_get_data(GTK_OBJECT(grp), "Label");
  g_return_if_fail(label != NULL);
  gtk_object_remove_data(GTK_OBJECT(tree), label);

  gtk_tree_remove_items(GTK_TREE(tree), l);
}

void hm_change_host(GtkWidget * item)
{
  xmlNodePtr h, sub;
  char *s;

  g_return_if_fail(item != NULL);

  h = (xmlNodePtr) gtk_object_get_data(GTK_OBJECT(item), "Node");
  g_return_if_fail(h != NULL);

  hm_clear_host();
  s = xmlGetProp(h, "name");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.sitename), s);
  g_free(s);
  gtk_entry_set_editable(GTK_ENTRY(hostmanager.sitename), FALSE);

  sub = h->childs;
  while (sub) {
    if (g_strcasecmp(sub->name, "hostname") == 0) {
      s = xmlNodeListGetString(doc, sub->childs, 1);
      gtk_entry_set_text(GTK_ENTRY(hostmanager.hostname), s ? s : "");
      if (s)
	g_free(s);
    } else if (g_strcasecmp(sub->name, "port") == 0) {
      s = xmlGetProp(sub, "value");
      gtk_entry_set_text(GTK_ENTRY(hostmanager.port), s ? s : "21");
      if (s)
	g_free(s);
    } else if (g_strcasecmp(sub->name, "anonymous") == 0) {
      s = xmlGetProp(sub, "value");
      if (atoi(s)) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hostmanager.anonymous), TRUE);
	gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.username), FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.password), FALSE);
      }
      g_free(s);
    } else if (g_strcasecmp(sub->name, "username") == 0) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hostmanager.anonymous), FALSE);
      gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.username), TRUE);
      gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.password), TRUE);
      s = xmlNodeListGetString(doc, sub->childs, 1);
      gtk_entry_set_text(GTK_ENTRY(hostmanager.username), s ? s : "");
      if (s)
	g_free(s);
    } else if (g_strcasecmp(sub->name, "password") == 0) {
      s = xmlNodeListGetString(doc, sub->childs, 1);
      gtk_entry_set_text(GTK_ENTRY(hostmanager.password), s ? s : "");
      if (s)
	g_free(s);
    } else if (g_strcasecmp(sub->name, "local_dir") == 0) {
      s = xmlNodeListGetString(doc, sub->childs, 1);
      gtk_entry_set_text(GTK_ENTRY(hostmanager.local_path), s ? s : "");
      if (s)
	g_free(s);
    } else if (g_strcasecmp(sub->name, "remote_dir") == 0) {
      s = xmlNodeListGetString(doc, sub->childs, 1);
      gtk_entry_set_text(GTK_ENTRY(hostmanager.remote_path), s ? s : "");
      if (s)
	g_free(s);
    } else if (g_strcasecmp(sub->name, "notes") == 0) {
      s = xmlNodeListGetString(doc, sub->childs, 1);
      gtk_text_insert(GTK_TEXT(hostmanager.notes), NULL, NULL, NULL, s ? s : "", -1);
      if (s)
	g_free(s);
    } else if (g_strcasecmp(sub->name, "timeout") == 0) {
      s = xmlGetProp(sub, "value");
      if (s) {
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(hostmanager.timeout),
				  (gfloat) atoi(s));
	g_free(s);
      }
    } else if (g_strcasecmp(sub->name, "retry_delay") == 0) {
      s = xmlGetProp(sub, "value");
      if (s) {
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(hostmanager.retry_delay),
				  (gfloat) atoi(s));
	g_free(s);
      }
    } else if (g_strcasecmp(sub->name, "retry_count") == 0) {
      s = xmlGetProp(sub, "value");
      if (s) {
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(hostmanager.retry_count),
				  (gfloat) atoi(s));
	g_free(s);
      }
    } else if (g_strcasecmp(sub->name, "show_hidden") == 0) {
      s = xmlGetProp(sub, "value");
      if (s) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hostmanager.show_hidden),
				     atoi(s));
	g_free(s);
      }
    } else if (g_strcasecmp(sub->name, "use_pasv") == 0) {
      s = xmlGetProp(sub, "value");
      if (s) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hostmanager.use_pasv),
				     atoi(s));
	g_free(s);
      }
    } else if (g_strcasecmp(sub->name, "keepalive") == 0) {
      s = xmlGetProp(sub, "value");
      if (s) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hostmanager.keepalive),
				     atoi(s));
	g_free(s);
      }
    }
    sub = sub->next;
  }
}
void hm_clear_host()
{
  gtk_entry_set_text(GTK_ENTRY(hostmanager.sitename), "");
  gtk_entry_set_editable(GTK_ENTRY(hostmanager.sitename), TRUE);
  gtk_entry_set_text(GTK_ENTRY(hostmanager.hostname), "");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.port), "21");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.username), "");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.password), "");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.local_path), "");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.remote_path), "");
  gtk_text_backward_delete(GTK_TEXT(hostmanager.notes),
			   gtk_text_get_length(GTK_TEXT(hostmanager.notes)));

  gtk_spin_button_set_value(GTK_SPIN_BUTTON(hostmanager.timeout),
			    (gfloat) deadftp_prefs.connect_timeout);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(hostmanager.retry_delay),
			    (gfloat) deadftp_prefs.retry_delay);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(hostmanager.retry_count),
			    (gfloat) deadftp_prefs.retry_count);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hostmanager.show_hidden),
			       deadftp_prefs.list_method);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hostmanager.use_pasv),
			       deadftp_prefs.firewall.use_pasv);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hostmanager.keepalive),
			       deadftp_prefs.anti_idle);

}
void hm_new_site(const char *name)
{
  if (!name || !*name) {
    return;
  }

  gtk_entry_set_text(GTK_ENTRY(hostmanager.sitename), name);
  gtk_entry_set_editable(GTK_ENTRY(hostmanager.sitename), FALSE);
  gtk_entry_set_text(GTK_ENTRY(hostmanager.hostname), "");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.port), "21");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.username), "");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.password), "");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.local_path), "");
  gtk_entry_set_text(GTK_ENTRY(hostmanager.remote_path), "");
  gtk_text_backward_delete(GTK_TEXT(hostmanager.notes),
			   gtk_text_get_length(GTK_TEXT(hostmanager.notes)));

}
static xmlNodePtr find_Host_Node(xmlNodePtr host, const char *n)
{
  xmlNodePtr node;

  g_return_val_if_fail(host != NULL, NULL);
  g_return_val_if_fail(n != NULL, NULL);

  node = host->childs;
  while (node) {
    if (g_strcasecmp(n, node->name) == 0)
      return node;
    node = node->next;
  }
  return NULL;
}
static void hm_check_and_update(xmlNodePtr host, const char *attr,
				const char *newvalue)
{
  xmlNodePtr node;
  char *oldvalue = NULL;

  g_return_if_fail(host != NULL);
  g_return_if_fail(attr != NULL);

  node = find_Host_Node(host, attr);
  if (node)
    oldvalue = xmlNodeListGetString(doc, node->childs, 1);
  if (!node && newvalue && *newvalue)
    xmlNewTextChild(host, ns, attr, newvalue);
  else if (g_strcasecmp(oldvalue ? oldvalue : "", newvalue ? newvalue : "") != 0) {
    xmlUnlinkNode(node);
    xmlFreeNode(node);
    if (newvalue && *newvalue)
      xmlNewTextChild(host, ns, attr, newvalue);
  }
  if (oldvalue)
    g_free(oldvalue);
}

void hm_change_host_info(GtkWidget * item)
{
  xmlNodePtr h, node;
  int i, j;
  char *value;

  g_return_if_fail(item != NULL);

  h = gtk_object_get_data(GTK_OBJECT(item), "Node");
  g_return_if_fail(h != NULL);

  hm_check_and_update(h, "hostname",
		      gtk_entry_get_text(GTK_ENTRY(hostmanager.hostname)));

  hm_check_and_update(h, "username",
		      gtk_entry_get_text(GTK_ENTRY(hostmanager.username)));

  hm_check_and_update(h, "password",
		      gtk_entry_get_text(GTK_ENTRY(hostmanager.password)));

  hm_check_and_update(h, "local_dir",
		      gtk_entry_get_text(GTK_ENTRY(hostmanager.local_path)));

  hm_check_and_update(h, "remote_dir",
		      gtk_entry_get_text(GTK_ENTRY(hostmanager.remote_path)));

  hm_check_and_update(h, "notes",
		      gtk_editable_get_chars(GTK_EDITABLE(hostmanager.notes), 0,
					     gtk_text_get_length(GTK_TEXT
								 (hostmanager.
								  notes))));

  node = find_Host_Node(h, "anonymous");
  value = xmlGetProp(node, "value");
  i = atoi(value) ? TRUE : FALSE;
  g_free(value);
  if (i != gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(hostmanager.anonymous)))
    xmlSetProp(node, "value", i ? "0" : "1");

  node = find_Host_Node(h, "port");
  value = xmlGetProp(node, "value");
  i = atoi(value);
  g_free(value);
  value = gtk_entry_get_text(GTK_ENTRY(hostmanager.port));
  j = atoi(value);
  if (i != j)
    xmlSetProp(node, "value", value);

  check_and_update_toggle(h, "timeout",
			  gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON
							   (hostmanager.timeout)));
  check_and_update_toggle(h, "retry_delay",
			  gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON
							   (hostmanager.
							    retry_delay)));
  check_and_update_toggle(h, "retry_count",
			  gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON
							   (hostmanager.
							    retry_count)));
  check_and_update_toggle(h, "show_hidden",
			  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON
						       (hostmanager.show_hidden)));
  check_and_update_toggle(h, "use_pasv",
			  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON
						       (hostmanager.use_pasv)));
  check_and_update_toggle(h, "keepalive",
			  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON
						       (hostmanager.keepalive)));
  /* TODO: Add code for resume options */
}
void check_and_update_toggle(xmlNodePtr hst, const char *attr, int newvalue)
{
  xmlNodePtr node;
  char buffer[SMALL_BUFFER], *o;
  int i;

  g_return_if_fail(hst != NULL);
  g_return_if_fail(attr != NULL);

  node = find_Host_Node(hst, attr);
  o = xmlGetProp(node, "value");
  if (o) {
    i = atoi(o);
    g_free(o);
  } else
    i = -1;
  if (i != newvalue) {
    if (!node)
      node = xmlNewChild(hst, ns, attr, NULL);
    g_snprintf(buffer, SMALL_BUFFER, "%d", newvalue);
    xmlSetProp(node, "value", buffer);
  }
}
void check_and_update_char_toggle(GtkWidget * item, const char *attr,
				  const char *newvalue)
{
  xmlNodePtr hst, node;
  char *oldvalue;

  g_return_if_fail(item != NULL);
  g_return_if_fail(attr != NULL);

  hst = gtk_object_get_data(GTK_OBJECT(item), "Node");
  g_return_if_fail(hst != NULL);

  node = find_Host_Node(hst, attr);
  oldvalue = xmlGetProp(node, "value");
  if (g_strcasecmp(oldvalue ? oldvalue : "", newvalue ? newvalue : "") != 0) {
    if (!node)
      node = xmlNewChild(hst, ns, attr, NULL);
    xmlSetProp(node, "value", newvalue);
  }
  if (oldvalue)
    g_free(oldvalue);
}

int hm_is_group(GtkWidget * item)
{
  xmlNodePtr node;

  g_return_val_if_fail(item != NULL, 0);

  node = gtk_object_get_data(GTK_OBJECT(item), "Node");
  g_return_val_if_fail(node != NULL, 0);

  if (g_strcasecmp(node->name, "HostGroup") == 0)
    return 1;
  return 0;
}
static void hm_ctree_add_host(GtkWidget * tree, GtkCTreeNode * parent,
			      xmlNodePtr node)
{
  gchar *row[3];
  xmlNodePtr sub;
  GtkCTreeNode *n;
  GdkPixmap *pixmap;
  GdkBitmap *mask;

  g_return_if_fail(tree != NULL);
  g_return_if_fail(node != NULL);

  load_gdk_pixmap(deadftp.window, "host.xpm", &pixmap, &mask);
  row[0] = xmlGetProp(node, "name");
  sub = find_Host_Node(node, "hostname");
  row[1] = xmlNodeListGetString(doc, sub->childs, 1);
  sub = find_Host_Node(node, "username");
  if (!sub)
    row[2] = strdup("anonymous");	/* strdup so we dont have to keep track when we free later */
  else
    row[2] = xmlNodeListGetString(doc, sub->childs, 1);
  n =
    gtk_ctree_insert_node(GTK_CTREE(tree), parent, NULL, row, 0, pixmap, mask, NULL,
			  NULL, TRUE, FALSE);
  gtk_ctree_node_set_row_data(GTK_CTREE(tree), n, (gpointer) node);

  g_free(row[0]);
  g_free(row[1]);
  g_free(row[2]);
  gdk_pixmap_unref(pixmap);
  gdk_bitmap_unref(mask);
}
static void hm_ctree_add_group(GtkWidget * tree, GtkCTreeNode * parent,
			       xmlNodePtr node)
{
  GtkCTreeNode *n;
  xmlNodePtr sub;
  gchar *row[3];
  GdkPixmap *pixmap;
  GdkBitmap *mask;

  g_return_if_fail(tree != NULL);
  g_return_if_fail(node != NULL);

  load_gdk_pixmap(deadftp.window, "hostgroup.xpm", &pixmap, &mask);
  row[0] = xmlGetProp(node, "name");
  row[1] = NULL;
  row[2] = NULL;
  n =
    gtk_ctree_insert_node(GTK_CTREE(tree), parent, NULL, row, 0, pixmap, mask,
			  pixmap, mask, FALSE, FALSE);
  gtk_ctree_node_set_row_data(GTK_CTREE(tree), n, (gpointer) node);
  sub = node->childs;
  while (sub) {
    if (g_strcasecmp(sub->name, "HostGroup") == 0)
      hm_ctree_add_group(tree, n, sub);
    else if (g_strcasecmp(sub->name, "Host") == 0)
      hm_ctree_add_host(tree, n, sub);
    sub = sub->next;
  }
  g_free(row[0]);
  gdk_pixmap_unref(pixmap);
  gdk_bitmap_unref(mask);
}

void hm_redisplay_ctree()
{
  xmlNodePtr node;

  gtk_clist_freeze(GTK_CLIST(hostmanager.host_ctree));
  gtk_clist_clear(GTK_CLIST(hostmanager.host_ctree));

  node = doc->root->childs;
  while (node) {
    if (g_strcasecmp(node->name, "HostGroup") == 0)
      hm_ctree_add_group(hostmanager.host_ctree, NULL, node);
    node = node->next;
  }
  gtk_clist_thaw(GTK_CLIST(hostmanager.host_ctree));
}

void hm_ctree_connect(GtkCTreeNode * selected_node)
{
  xmlNodePtr node, sub;
  char *val1, *val2;

  g_return_if_fail(selected_node != NULL);
  node =
    (xmlNodePtr) gtk_ctree_node_get_row_data(GTK_CTREE(hostmanager.host_ctree),
					     selected_node);
  g_return_if_fail(node != NULL);

  g_return_if_fail(g_strcasecmp(node->name, "Host") == 0);

  sub = find_Host_Node(node, "hostname");
  val1 = xmlNodeListGetString(doc, sub->childs, 1);
  strncpy(FTP.server.hostname, val1, SMALL_BUFFER);
  g_free(val1);
  sub = find_Host_Node(node, "port");
  val1 = xmlGetProp(sub, "value");
  FTP.server.port = atoi(val1);
  g_free(val1);
  sub = find_Host_Node(node, "anonymous");
  val1 = xmlGetProp(sub, "value");
  if (atoi(val1)) {
    strncpy(FTP.server.username, "anonymous", SMALL_BUFFER);
    strncpy(FTP.server.password, deadftp_prefs.anonymous_password, SMALL_BUFFER);
  } else {
    sub = find_Host_Node(node, "username");
    val2 = xmlNodeListGetString(doc, sub->childs, 1);
    strncpy(FTP.server.username, val2, SMALL_BUFFER);
    g_free(val2);
    sub = find_Host_Node(node, "password");
    val2 = xmlNodeListGetString(doc, sub->childs, 1);
    strncpy(FTP.server.password, val2, SMALL_BUFFER);
    g_free(val2);
  }
  g_free(val1);
  sub = find_Host_Node(node, "local_dir");
  if (sub) {
    val1 = xmlNodeListGetString(doc, sub->childs, 1);
    chdir(val1);
    g_free(val1);
    change_local();
  }
  sub = find_Host_Node(node, "remote_dir");
  if (sub) {
    val1 = xmlNodeListGetString(doc, sub->childs, 1);
    strncpy(FTP.remote_dir, val1, SMALL_BUFFER);
    g_free(val1);
    SET_FLAG(FTP_FLAG_CHANGE_DIR);
  }

  FTP.host_entry = (gpointer) node;
  ftp_connect();
}

void hm_show_ctree()
{
  if (ctree_need_redraw)
    hm_redisplay_ctree();

  gtk_widget_hide(deadftp.remote_scroll);
  gtk_widget_show(deadftp.host_scroll);

}

void hm_hide_ctree()
{
  gtk_widget_hide(deadftp.host_scroll);
  gtk_widget_show(deadftp.remote_scroll);

}

void on_add_host_clicked(GtkButton * button, gpointer user_data)
{
  GtkWidget *w;
  char *site_val, *host_val, *port_val, *user_val, *pass_val, *local_val,
    *remote_val, *notes_val;
  int timeout_val, delay_val, retry_val, show_hidden_val, use_pasv_val,
    keepalive_val;
  GList *sel;

  sel = GTK_TREE_SELECTION(GTK_TREE(hostmanager.host_tree));
  w = g_list_nth_data(sel, 0);
  if (!w) {
    w = gtk_object_get_data(GTK_OBJECT(hostmanager.host_tree), "General");
  }

  if (!w) {
    g_print(_("No Host Groups to add host to!\n"));
    return;
  }
  if (gtk_object_get_data(GTK_OBJECT(w), "HostGroup")) {
    w = gtk_object_get_data(GTK_OBJECT(w), "HostGroup");
  }

  site_val = gtk_entry_get_text(GTK_ENTRY(hostmanager.sitename));
  host_val = gtk_entry_get_text(GTK_ENTRY(hostmanager.hostname));
  port_val = gtk_entry_get_text(GTK_ENTRY(hostmanager.port));
  user_val = gtk_entry_get_text(GTK_ENTRY(hostmanager.username));
  pass_val = gtk_entry_get_text(GTK_ENTRY(hostmanager.password));
  local_val = gtk_entry_get_text(GTK_ENTRY(hostmanager.local_path));
  remote_val = gtk_entry_get_text(GTK_ENTRY(hostmanager.remote_path));
  notes_val =
    gtk_editable_get_chars(GTK_EDITABLE(hostmanager.notes), 0,
			   gtk_text_get_length(GTK_TEXT(hostmanager.notes)));

  timeout_val =
    gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(hostmanager.timeout));
  delay_val =
    gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(hostmanager.retry_delay));
  retry_val =
    gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(hostmanager.retry_count));

  show_hidden_val =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(hostmanager.show_hidden));
  use_pasv_val =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(hostmanager.use_pasv));
  keepalive_val =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(hostmanager.keepalive));

  /* TODO: resume values */

  hm_add_host(gtk_object_get_data(GTK_OBJECT(w), "Label"), site_val, host_val,
	      port_val, user_val,
	      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(hostmanager.anonymous)),
	      pass_val, local_val, remote_val, notes_val, timeout_val, delay_val,
	      retry_val, show_hidden_val, use_pasv_val, keepalive_val);

  hostmanager_changed();
}

void on_add_group_clicked(GtkButton * button, gpointer user_data)
{
  GladeXML *xml;

  xml = glade_xml_new(deadftp.gladefile, "groupnameDialog");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  hostmanager.groupnameDialog.dialog = glade_xml_get_widget(xml, "groupnameDialog");
  STORE_WIDGET(hostmanager.groupnameDialog, group_name);

  glade_xml_signal_autoconnect(xml);

  gtk_object_unref(GTK_OBJECT(xml));
}

void on_group_ok_clicked(GtkButton * button, gpointer user_data)
{
  hm_add_group(gtk_entry_get_text
	       (GTK_ENTRY(hostmanager.groupnameDialog.group_name)));
  gtk_widget_destroy(hostmanager.groupnameDialog.dialog);

  hostmanager_changed();
}

void on_group_cancel_clicked(GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy(hostmanager.groupnameDialog.dialog);
}

void on_hm_connect_clicked(GtkButton * button, gpointer user_data)
{
  char *p;
  GList *l;
  GtkWidget *item;

  strncpy(FTP.server.hostname, gtk_entry_get_text(GTK_ENTRY(hostmanager.hostname)),
	  SMALL_BUFFER);
  if ((p = gtk_entry_get_text(GTK_ENTRY(hostmanager.port)))) {
    FTP.server.port = atoi(p);
  } else
    FTP.server.port = 21;
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(hostmanager.anonymous))) {
    strncpy(FTP.server.username, "anonymous", SMALL_BUFFER);
    strncpy(FTP.server.password, deadftp_prefs.anonymous_password, SMALL_BUFFER);
  } else {
    strncpy(FTP.server.username, gtk_entry_get_text(GTK_ENTRY(hostmanager.username)),
	    SMALL_BUFFER);
    strncpy(FTP.server.password, gtk_entry_get_text(GTK_ENTRY(hostmanager.password)),
	    SMALL_BUFFER);
  }
  p = gtk_entry_get_text(GTK_ENTRY(hostmanager.local_path));
  if (p) {
    chdir(p);
    change_local();
  }
  p = gtk_entry_get_text(GTK_ENTRY(hostmanager.remote_path));
  if (p) {
    strncpy(FTP.remote_dir, gtk_entry_get_text(GTK_ENTRY(hostmanager.remote_path)),
	    SMALL_BUFFER);
    SET_FLAG(FTP_FLAG_CHANGE_DIR);
  } else {
    CLEAR_FLAG(FTP_FLAG_CHANGE_DIR);
  }

  l = GTK_TREE_SELECTION(hostmanager.host_tree);
  item = g_list_nth_data(l, 0);

  FTP.host_entry = gtk_object_get_data(GTK_OBJECT(item), "Node");

  delete_hostmanager();

  ftp_connect();
}

void on_hm_exit_clicked(GtkButton * button, gpointer user_data)
{
  delete_hostmanager();
}

void on_hm_apply_clicked(GtkButton * button, gpointer user_data)
{
  GtkWidget *item;
  GList *l;

  l = GTK_TREE_SELECTION(GTK_TREE(hostmanager.host_tree));
  item = g_list_nth_data(l, 0);
  if (!item)			/* New Host */
    on_add_host_clicked(button, user_data);
  else {
    hm_change_host_info(item);
  }
  hostmanager.changed = FALSE;
  gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.hm_apply), FALSE);
}

void on_anonymous_toggled(GtkToggleButton * togglebutton, gpointer user_data)
{
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(togglebutton))) {
    gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.username), FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.password), FALSE);
  } else {
    gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.username), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.password), TRUE);
  }
  hostmanager_changed();
}

void on_delete_host_clicked(GtkButton * button, gpointer user_data)
{
  GtkWidget *item;
  GList *selection;

  if (!GTK_TREE_SELECTION(GTK_TREE(hostmanager.host_tree)))
    return;
  selection = GTK_TREE_SELECTION(GTK_TREE(hostmanager.host_tree));
  item = (GtkWidget *) g_list_nth_data(selection, 0);
  if (item) {
    GtkWidget *tree = gtk_object_get_data(GTK_OBJECT(item), "HostGroup");

    g_return_if_fail(tree != NULL);
    hm_delete_host(tree, item);
    hostmanager_changed();
  }
}
void on_delete_group_clicked(GtkButton * button, gpointer user_data)
{
  GtkWidget *item;
  GList *selection;

  if (!GTK_TREE_SELECTION(GTK_TREE(hostmanager.host_tree)))
    return;
  selection = GTK_TREE_SELECTION(GTK_TREE(hostmanager.host_tree));
  item = (GtkWidget *) g_list_nth_data(selection, 0);
  if (item) {
    if (gtk_object_get_data(GTK_OBJECT(item), "HostGroup"))
      return;
    hm_delete_group(item);
    hostmanager_changed();
  }
}
void on_host_tree_select_child(GtkTree * tree, GtkWidget * widget,
			       gpointer user_data)
{
  GtkWidget *grp = NULL;
  gpointer data;
  gboolean changed;

  /* Store the changed value since changing the host information will trigger
   * hostmanager_changed() even if nothing has been changed by the user */
  changed = hostmanager.changed;

  data = gtk_object_get_data(GTK_OBJECT(widget), "HostGroup");
  if (!data) {
    hm_clear_host();
    return;
  }
  if (hm_is_group(widget)) {
    hm_clear_host();
    return;
  }
  grp = GTK_WIDGET(data);
  if (grp)
    hm_change_host(widget);
  else
    hm_clear_host();

  if (!changed) {
    hostmanager.changed = FALSE;
    gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.hm_apply), FALSE);
  }
}

void on_hm_new_clicked(GtkButton * button, gpointer user_data)
{
  GladeXML *xml;

  xml = glade_xml_new(deadftp.gladefile, "sitenameDialog");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  hostmanager.sitenameDialog.dialog = glade_xml_get_widget(xml, "sitenameDialog");
  STORE_WIDGET(hostmanager.sitenameDialog, sitename);

  glade_xml_signal_autoconnect(xml);

  gtk_object_unref(GTK_OBJECT(xml));
}

void on_sitename_ok_clicked(GtkButton * button, gpointer user_data)
{
  char port[32], cwd[SMALL_BUFFER], *name;

  if (gtk_object_get_data
      (GTK_OBJECT(hostmanager.sitenameDialog.dialog), "add_current")) {
    g_snprintf(port, 32, "%d", FTP.server.port);
    getcwd(cwd, SMALL_BUFFER);
    name = gtk_entry_get_text(GTK_ENTRY(hostmanager.sitenameDialog.sitename));
    hm_add_host("General", name, FTP.server.hostname, port, FTP.server.username,
		(g_strcasecmp(FTP.server.username, "anonymous") == 0) ? 1 : 0,
		FTP.server.password, cwd, FTP.server.current_directory, NULL,
		deadftp_prefs.connect_timeout, deadftp_prefs.retry_delay,
		deadftp_prefs.retry_count, deadftp_prefs.list_method,
		deadftp_prefs.firewall.use_pasv, deadftp_prefs.anti_idle);
  } else {
    name = gtk_entry_get_text(GTK_ENTRY(hostmanager.sitenameDialog.sitename));
    hm_new_site(name);
  }
  gtk_widget_destroy(hostmanager.sitenameDialog.dialog);
  hostmanager_changed();
}

void on_sitename_cancel_clicked(GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy(hostmanager.sitenameDialog.dialog);
}

void on_hostmanager_destroy(GtkObject * object, gpointer user_data)
{
  hostmanager_is_gone();
}

gboolean on_host_ctree_button_press_event(GtkWidget * widget, GdkEventButton * event,
					  gpointer user_data)
{
  if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1))
    hm_ctree_connect(ctree_selected_node);
  return FALSE;
}

void on_host_ctree_tree_select_row(GtkCTree * ctree, GList * node, gint column,
				   gpointer user_data)
{
  ctree_selected_node = (GtkCTreeNode *) node;
}

void on_add_cur_host_clicked(GtkButton * button, gpointer user_data)
{
  GladeXML *xml;

  if (FTP.sock->sockfd == -1)
    return;

  xml = glade_xml_new(deadftp.gladefile, "sitenameDialog");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  hostmanager.sitenameDialog.dialog = glade_xml_get_widget(xml, "sitenameDialog");
  STORE_WIDGET(hostmanager.sitenameDialog, sitename);

  glade_xml_signal_autoconnect(xml);

  gtk_object_unref(GTK_OBJECT(xml));

  /* the data that's set doesn't matter, the key just has to exist for test */
  gtk_object_set_data(GTK_OBJECT(hostmanager.sitenameDialog.dialog), "add_current",
		      button);

}

void on_hostmanag_clicked(GtkButton * button, gpointer user_data)
{
  start_hostmanager();
}

void on_hostmanager1_activate(GtkMenuItem * menuitem, gpointer user_data)
{
  start_hostmanager();
}

int hm_get_pref_int(char *attr)
{
  xmlNodePtr node, value;
  char *val;
  int ret;

  if (!FTP.host_entry)
    return -2;
  node = (xmlNodePtr) FTP.host_entry;
  value = find_Host_Node(node, attr);

  val = xmlGetProp(value, "value");
  if (!val) {
    ret = -2;
  } else {
    ret = atoi(val);
    g_free(val);
  }

  return ret;
}

void hostmanager_changed(void)
{
  hostmanager.changed = TRUE;

  gtk_widget_set_sensitive(GTK_WIDGET(hostmanager.hm_apply), TRUE);
}
