/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include "globals.h"
#include "main.h"
#include "local.h"
#include "ftp.h"
#include "layout.h"
#include "support.h"

/* Callbacks in this file */
gboolean on_local_button_press_event(GtkWidget * widget, GdkEventButton * event,
				     gpointer user_data);
void on_local_textcombo_activate(GtkEditable * editable, gpointer user_data);
void on_local_select_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			 gpointer user_data);
void on_local_drag_data_get(GtkWidget * widget, GdkDragContext * drag_context,
			    GtkSelectionData * data, guint info, guint time,
			    gpointer user_data);
void on_local_up_clicked(GtkButton * button, gpointer user_data);
void on_local_unselect_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			   gpointer user_data);
void on_local_unselect_all(GtkCList * clist, gpointer user_data);
void on_dir_ok_clicked(GtkButton * button, gpointer user_data);
void on_dir_cancel_clicked(GtkButton * button, gpointer user_data);
void on_rename_ok_clicked(GtkButton * button, gpointer user_data);
void on_rename_cancel_clicked(GtkButton * button, gpointer user_data);
void on_local_click_column(GtkCList * clist, gint column);
void on_local_arrange_name(void);
void on_local_arrange_length(void);
void on_local_arrange_mode(void);
void on_local_arrange_owner(void);
void on_local_arrange_group(void);
void on_local_arrange_date(void);
void on_local_arrange_ascending(void);
void on_local_arrange_descending(void);
void on_local_arrange_autosize(GtkMenuItem *);
void on_local_move(void);
void on_local_mode(void);
void on_local_transfer(void);
void on_mode_toggled(GtkToggleButton * button, gpointer data);
void on_mode_changed(GtkEntry * entry, gpointer data);
void on_mode_ok_clicked(GtkButton * button, gpointer data);
void on_local_mkdir(void);
void on_local_remove(void);
void on_local_rename(void);
void on_local_add_queue(void);
void on_local_refresh(void);


/* Functions local to this file */
static void change_local_combo(void);
static void change_local_dir(void);
static void local_remove_dir(const char *dir);
static char *mode_to_str(mode_t mode);
static void local_dblclick(void);
static char *uid_to_str(uid_t uid);
static char *gid_to_str(gid_t gid);
void local_do_rename(const char *, const char *);
gint sort_file(gconstpointer a, gconstpointer b);
static gint sort_owner(gconstpointer a, gconstpointer b);
static gint sort_group(gconstpointer a, gconstpointer b);
static gint sort_mode(gconstpointer a, gconstpointer b);
static gint sort_length(gconstpointer a, gconstpointer b);
static gint sort_date(gconstpointer a, gconstpointer b);
static int month_to_int(const char *month);
static void local_set_sort_info(gint, gboolean);
static void local_move_string_callback(gchar * string, gpointer data);
char *find_pixmap_file(const char *);

const char *dir_layout_titles[] = {
  N_("Filename"),
  N_("Length"),
  N_("Mode"),
  N_("Owner"),
  N_("Group"),
  N_("Date")
};

const struct sort_func dir_sort_funcs[] = {
  {sort_file},
  {sort_length},
  {sort_mode},
  {sort_owner},
  {sort_group},
  {sort_date},
};

extern GList *remote_files;

extern GMemChunk *afile_chunk;
GStringChunk *afile_local_string_chunk = NULL;

GList *local_row = NULL;
GList *local_files = NULL;

void change_local()
{
  change_local_combo();
  change_local_dir();
}
static void change_local_combo()
{
  char cwd[SMALL_BUFFER], *ch;
  GList *l, *strings;

  getcwd(cwd, SMALL_BUFFER);
  ch = strrchr(cwd, '/');
  strings = g_list_append(NULL, g_strdup(cwd));
  while (ch) {
    *ch = '\0';
    strings = g_list_append(strings, g_strdup(cwd));
    ch = strrchr(cwd, '/');
  }
  gtk_combo_set_popdown_strings(GTK_COMBO(deadftp.local_combo), strings);

  l = strings;
  while (l) {
    ch = (char *) l->data;
    g_free(ch);
    l = g_list_next(l);
  }
  g_list_free(l);

}
static void change_local_dir()
{
  GList *g = NULL;
  DIR *dir;
  struct stat sbuf;
  char buf[SMALL_BUFFER], *cwd;
  struct dirent *d;
  struct afile *file, *x;
  struct tm mod, now;
  time_t n;

  cwd = getcwd(NULL, 0);
  dir = opendir(cwd);
  g_free(cwd);
  if (!dir) {
    return;
  }

  n = time(NULL);
  localtime_r(&n, &now);

  for (g = local_files; g; g = g_list_next(g)) {
    x = (struct afile *) g->data;
    g_chunk_free(x, afile_chunk);
  }
  g_list_free(local_files);
  local_files = NULL;
  /* free strings if they exist */
  if (afile_local_string_chunk)
    g_string_chunk_free(afile_local_string_chunk);

  afile_local_string_chunk = g_string_chunk_new(1024);
  while ((d = readdir(dir))) {
    if (!deadftp_prefs.list_method == LIST_SHOW_HIDDEN && *d->d_name == '.') {
      /* This is a hidden file and show_hidden is off */
      continue;
    }
    file = g_chunk_new0(struct afile, afile_chunk);
    file->file = g_string_chunk_insert_const(afile_local_string_chunk, d->d_name);
    if (stat(d->d_name, &sbuf) < 0)
      fprintf(stderr, _("couldn't stat %s\n"), d->d_name);

    g_snprintf(buf, 32, "%d", (int) sbuf.st_size);
    file->length = g_string_chunk_insert_const(afile_local_string_chunk, buf);

    file->mode = mode_to_str(sbuf.st_mode);

    file->owner = uid_to_str(sbuf.st_uid);

    file->group = gid_to_str(sbuf.st_gid);

    localtime_r(&sbuf.st_mtime, &mod);
    /* If the Modification date is older than 6 months then print year instead of time */
    if ((now.tm_year == mod.tm_year) && ((now.tm_mon - mod.tm_mon) < 6)) {
      strftime(buf, SMALL_BUFFER, "%b %d %H:%M", &mod);
    } else {
      strftime(buf, SMALL_BUFFER, "%b %d %Y", &mod);
    }
    file->date = g_string_chunk_insert_const(afile_local_string_chunk, buf);

    if (sbuf.st_mode & S_IFDIR)
      file->is_dir = 1;
    else
      file->is_dir = 0;
    local_files =
      g_list_insert_sorted(local_files, file,
			   dir_sort_funcs[deadftp_prefs.local_sort_field].func);
  }

  local_refresh_dir();

  closedir(dir);
}

void local_refresh_dir()
{
  int i, j;
  char *dirent[NUM_DIR_LAYOUTS];
  struct afile *x;

  /* local_files = g_list_sort(local_files, dir_sort_funcs[deadftp_prefs.local_sort_field].func); */
  if (deadftp_prefs.local_sort_order == DIR_SORT_DSC) {
    local_files = g_list_reverse(local_files);
  }

  g_list_free(local_row);
  local_row = NULL;

  gtk_clist_clear(GTK_CLIST(deadftp.local));
  gtk_clist_freeze(GTK_CLIST(deadftp.local));

  for (i = 0; i < g_list_length(local_files); i++) {
    x = g_list_nth_data(local_files, i);

    for (j = 0; j < NUM_DIR_LAYOUTS; j++)
      dirent[j] = NULL;

    if (deadftp_prefs.directory_layout[DIR_LAYOUT_FILE] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_FILE]] = x->file;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH] != -1) {
      char xtra[8], length[SMALL_BUFFER];
      int len = atoi(x->length);

      xtra[0] = '\0';
      if (len > (ONE_KILOBYTE * 10)) {
	len /= ONE_KILOBYTE;
	strncpy(xtra, "KB", 8);
      }
      if (len > (ONE_KILOBYTE * 10)) {
	len /= ONE_KILOBYTE;
	strncpy(xtra, "MB", 8);
      }

      g_snprintf(length, SMALL_BUFFER, "%d%s%s", len, (*xtra != '\0') ? " " : "",
		 xtra);

      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH]] = strdup(length);
    }
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_MODE] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_MODE]] = x->mode;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER]] = x->owner;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP]] = x->group;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_DATE] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_DATE]] = x->date;

    gtk_clist_append(GTK_CLIST(deadftp.local), dirent);
    gtk_clist_set_row_data(GTK_CLIST(deadftp.local), i, x);

    set_clist_icon(deadftp.local, i, deadftp_prefs.directory_layout[DIR_LAYOUT_FILE],
		   x->file, x->is_dir);

    if (deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH] != -1) {
      g_free(dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH]]);
    }
  }
  gtk_clist_thaw(GTK_CLIST(deadftp.local));
}

static char *mode_to_str(mode_t mode)
{
  char buffer[11];

  buffer[0] = '-';
  if (mode & S_IFDIR)
    buffer[0] = 'd';
  (mode & S_IRUSR) ? buffer[1] = 'r' : (buffer[1] = '-');
  (mode & S_IWUSR) ? buffer[2] = 'w' : (buffer[2] = '-');
  (mode & S_IXUSR) ? buffer[3] = 'x' : (buffer[3] = '-');
  if (mode & S_ISUID)
    buffer[3] = 's';
  (mode & S_IRGRP) ? buffer[4] = 'r' : (buffer[4] = '-');
  (mode & S_IWGRP) ? buffer[5] = 'w' : (buffer[5] = '-');
  (mode & S_IXGRP) ? buffer[6] = 'x' : (buffer[6] = '-');
  if (mode & S_ISGID)
    buffer[6] = 's';
  (mode & S_IROTH) ? buffer[7] = 'r' : (buffer[7] = '-');
  (mode & S_IWOTH) ? buffer[8] = 'w' : (buffer[8] = '-');
  (mode & S_IXOTH) ? buffer[9] = 'x' : (buffer[9] = '-');
  if (mode & S_ISVTX)
    buffer[9] = 't';
  buffer[10] = '\0';

  return g_string_chunk_insert_const(afile_local_string_chunk, buffer);
}

static char *uid_to_str(uid_t uid)
{
  /* Make pw static as a wuick hack to cache previous value */
  static struct passwd *pw = NULL;

  if (!pw || (pw->pw_uid != uid))
    pw = getpwuid(uid);

  if (!pw) {
    char uid_str[SMALL_BUFFER];

    g_snprintf(uid_str, SMALL_BUFFER, "%d", uid);
    return g_string_chunk_insert_const(afile_local_string_chunk, uid_str);
  }
  return g_string_chunk_insert_const(afile_local_string_chunk, pw->pw_name);
}

static char *gid_to_str(gid_t gid)
{
  /* Make gr static as a quick hack to cache the previous value to 
   * try and reduce the number of getgrgid calls we need to make */
  static struct group *gr = NULL;

  if (!gr || (gr->gr_gid != gid))
    gr = getgrgid(gid);

  if (!gr) {
    char gid_str[SMALL_BUFFER];

    g_snprintf(gid_str, SMALL_BUFFER, "%d", gid);
    return g_string_chunk_insert_const(afile_local_string_chunk, gid_str);
  }
  return g_string_chunk_insert_const(afile_local_string_chunk, gr->gr_name);
}

gint sort_file(gconstpointer a, gconstpointer b)
{
  struct afile *x, *y;

  x = (struct afile *) a;
  y = (struct afile *) b;
  /* when one is NULL use pointer subtraction to figure out which is greater */
  if (!x || !y) {
    return y - x;
  }
  if (!x->file || !y->file)
    return y->file - x->file;

  if (y->is_dir - x->is_dir != 0)
    return y->is_dir - x->is_dir;
  else
    return strcmp(x->file, y->file);
}

static gint sort_owner(gconstpointer a, gconstpointer b)
{
  struct afile *x, *y;
  char *str1, *str2;

  x = (struct afile *) a;
  y = (struct afile *) b;
  if (!x || !y) {
    return y - x;
  }
  if (!x->owner || !y->owner)
    return y->owner - x->owner;

  str1 = x->owner;
  str2 = y->owner;
  if (g_strcasecmp(str1, str2) == 0) {
    /* for rows with the same value, sort by filename */
    return sort_file(a, b);
  }
  return strcmp(str1, str2);
}

static gint sort_group(gconstpointer a, gconstpointer b)
{
  struct afile *x, *y;
  char *str1, *str2;

  x = (struct afile *) a;
  y = (struct afile *) b;
  if (!x || !y) {
    return y - x;
  }
  if (!x->group || !y->group)
    return y->group - x->group;

  str1 = x->group;
  str2 = y->group;
  if (g_strcasecmp(str1, str2) == 0) {
    /* for rows with the same value, sort by filename */
    return sort_file(a, b);
  }
  return strcmp(str1, str2);
}

static gint sort_mode(gconstpointer a, gconstpointer b)
{
  struct afile *x, *y;
  char *str1, *str2;

  x = (struct afile *) a;
  y = (struct afile *) b;
  if (!x || !y) {
    return y - x;
  }
  if (!x->mode || !y->mode)
    return y->mode - x->mode;

  str1 = x->mode;
  str2 = y->mode;
  if (g_strcasecmp(str1, str2) == 0) {
    /* for rows with the same value, sort by filename */
    return sort_file(a, b);
  }
  return strcmp(str1, str2);
}


static gint sort_length(gconstpointer a, gconstpointer b)
{
  struct afile *x, *y;
  int val1, val2;

  x = (struct afile *) a;
  y = (struct afile *) b;
  if (!x || !y) {
    return y - x;
  }
  if (!x->length || !y->length)
    return y->length - x->length;

  val1 = atoi(x->length);
  val2 = atoi(y->length);

  if (val1 == val2) {
    return sort_file(a, b);
  }
  return val2 - val1;
}

static gint sort_date(gconstpointer a, gconstpointer b)
{
  struct afile *x, *y;
  char mo1[4], mo2[4];
  int y1, y2, h1, h2, m1, m2, d1, d2, mon1, mon2;
  int ms1, ms2;

  x = (struct afile *) a;
  y = (struct afile *) b;
  if (!x || !y) {
    return y - x;
  }
  if (!x->date || !y->date)
    return y->date - x->date;

  ms1 = sscanf(x->date, "%s %d %d:%d", mo1, &d1, &h1, &m1);
  ms2 = sscanf(y->date, "%s %d %d:%d", mo2, &d2, &h2, &m2);
  if (ms1 != ms2) {
    /* One is a year format and the other is a hour:minute format */
    /* year format is older and will be 3 hour:minute is younger and will be 4 */
    return ms2 - ms1;
  }
  if (ms1 == 4) {
    /* If ms1 == 4 then ms2 must == 4 */
    mon1 = month_to_int(mo1);
    mon2 = month_to_int(mo2);
    if (mon1 != mon2) {
      return mon2 - mon1;
    }
    /* If we get here mon1 == mon2 */
    if (d1 != d2) {
      return d2 - d1;
    }
    /* If we get here d1 == d2 */
    m1 += h1 * 60;
    m2 += h2 * 60;
    return m2 - m1;
  }
  /* If we get here then either the date both dates are in Month day year format or an unkown format */
  ms1 = sscanf(x->date, "%s %d %d", mo1, &d1, &y1);
  ms2 = sscanf(y->date, "%s %d %d", mo2, &d2, &y2);
  if ((ms1 != 3) || (ms2 != 3)) {
    /* I don't know how to deal with these dates */
    return 0;
  }
  if (y1 != y2) {
    return y2 - y1;
  }
  /* y1 == y2 */
  mon1 = month_to_int(mo1);
  mon2 = month_to_int(mo2);
  if (mon1 != mon2) {
    return mon2 - mon1;
  }
  /* mon1 == mon2 */
  return d2 - d1;
}

static int month_to_int(const char *month)
{
  const char *months[] =
    { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov",
    "Dec"
  };
  int i;

  g_return_val_if_fail(month != NULL, -1);

  for (i = 0; i < 12; i++) {
    if (g_strcasecmp(month, months[i]) == 0) {
      return i;
    }
  }
  return -1;
}

static void local_dblclick()
{
  char cwd[REGULAR_BUFFER];
  struct afile *f;
  gint src;

  if (!local_row) {
    return;
  }
  src = GPOINTER_TO_INT(local_row->data);

  f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.local), src);
  g_return_if_fail(f != NULL);

  if (f->is_dir) {
    chdir(f->file);
    change_local();
  } else if (FTP.sock->sockfd > -1) {
    getcwd(cwd, REGULAR_BUFFER);
    add_queue(&FTP.server, f, cwd, f->file, TRUE);
  }
}

void load_newDir_dialog()
{
  GladeXML *xml;

  /* Do we need to be able to open more than one newDir dialog at a time?
   * If so, remove this line and restructure how the dialog is loaded so
   * information isn't stored in only 1 struct */
  g_return_if_fail(deadftp.newDir.dialog != NULL);

  xml = glade_xml_new(deadftp.gladefile, "newDir");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  deadftp.newDir.dialog = glade_xml_get_widget(xml, "newDir");

  STORE_WIDGET(deadftp.newDir, directory_name);

  glade_xml_signal_autoconnect(xml);
  gtk_object_unref(GTK_OBJECT(xml));
}

void on_local_mkdir()
{
  load_newDir_dialog();

  gtk_object_set_data(GTK_OBJECT(deadftp.newDir.dialog), "isLocal",
		      GINT_TO_POINTER(1));
}
static void local_remove_dir(const char *dir)
{
  DIR *dirp;
  struct dirent *d;
  struct stat sbuf;
  char newdir[REGULAR_BUFFER];
  static int subdir = 0;	/* we'll only go down 5 dirs or so, for now */

  g_return_if_fail(dir != NULL);

  subdir++;
  if (subdir > 5) {
    subdir--;
    return;
  }
  dirp = opendir(dir);
  if (!dirp) {
    subdir--;
    return;
  }
  while ((d = readdir(dirp))) {
    if ((strcmp(d->d_name, ".") == 0) || (strcmp(d->d_name, "..") == 0))
      continue;
    g_snprintf(newdir, REGULAR_BUFFER, "%s/%s", dir, d->d_name);
    if (stat(newdir, &sbuf) < 0)
      continue;			/* can't stat then we probably can't do anything else to it */
    if (sbuf.st_mode & S_IFDIR) {
      g_snprintf(newdir, REGULAR_BUFFER, "%s/%s", dir, d->d_name);
      local_remove_dir(newdir);
      continue;
    }
    unlink(newdir);
  }
  rmdir(dir);
  closedir(dirp);
}

void on_local_remove()
{
  GList *r, *copy;
  gint src;
  struct afile *f;

  copy = g_list_copy(local_row);
  for (r = copy; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.local), src);
    g_return_if_fail(f != NULL);

    if (f->is_dir) {
      local_remove_dir(f->file);	/* This will go remove recursively */
      r = g_list_next(r);
      continue;
    }
    /* does unlink need full path?? */
    unlink(f->file);
    /* need to check for error and bring up a dialog on failure */
  }
  gtk_clist_unselect_all(GTK_CLIST(deadftp.local));
  g_list_free(local_row);
  local_row = NULL;
  g_list_free(copy);
  change_local();
}

void on_local_rename()
{
  /* TODO: have this funtion bring up dialog for each of the selected files */
  int src;
  struct afile *f;

  if (!local_row)
    return;

  src = GPOINTER_TO_INT(local_row->data);

  f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.local), src);
  g_return_if_fail(f != NULL);

  load_rename_dialog();
  gtk_label_set_text(GTK_LABEL(deadftp.renameDialog.filename), f->file);
  gtk_object_set_data(GTK_OBJECT(deadftp.renameDialog.dialog), "local_filename",
		      f->file);
}

void on_local_add_queue()
{
  char buffer[REGULAR_BUFFER];
  GList *r;
  gint src;
  struct afile *f;

  if (!local_row) {
    return;
  }
  for (r = local_row; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.local), src);
    g_return_if_fail(f != NULL);

    getcwd(buffer, REGULAR_BUFFER);
    add_queue(&FTP.server, f, buffer, f->file, TRUE);
  }
}

GtkWidget *load_pixmap(GtkWidget * widget, const gchar * filename,
		       gboolean gnome_pixmap)
{
  GtkWidget *pixmap;
  GdkColormap *colormap;
  GdkPixmap *gdkpixmap;
  GdkBitmap *mask;
  gchar pathname[REGULAR_BUFFER], *dirname;

  if (!filename || !filename[0])
    return NULL;

  dirname = g_dirname(deadftp.gladefile);
  g_snprintf(pathname, REGULAR_BUFFER, "%s/%s", dirname, filename);
  g_free(dirname);

  if (!g_file_exists(pathname)) {
    g_warning(_("Couldn't find pixmap file: %s"), filename);
    return NULL;
  }

  if (gnome_pixmap) {
    pixmap = gnome_pixmap_new_from_file(pathname);
    return pixmap;
  }

  colormap = gtk_widget_get_colormap(widget);
  gdkpixmap =
    gdk_pixmap_colormap_create_from_xpm(NULL, colormap, &mask, NULL, pathname);
  if (gdkpixmap == NULL) {
    g_warning(_("Couldn't create pixmap from file: %s"), pathname);
    return NULL;
  }

  pixmap = gtk_pixmap_new(gdkpixmap, mask);
  gdk_pixmap_unref(gdkpixmap);
  gdk_bitmap_unref(mask);
  return pixmap;
}

void load_gdk_pixmap(GtkWidget * widget, const gchar * filename,
		     GdkPixmap ** gdkpixmap, GdkBitmap ** mask)
{
  GdkColormap *colormap;
  gchar pathname[REGULAR_BUFFER], *dirname;

  if (!filename || !filename[0])
    return;

  dirname = g_dirname(deadftp.gladefile);
  g_snprintf(pathname, REGULAR_BUFFER, "%s/%s", dirname, filename);
  g_free(dirname);
  if (!g_file_exists(pathname)) {
    g_warning(_("Couldn't find pixmap file: %s"), filename);
    return;
  }

  colormap = gtk_widget_get_colormap(widget);
  *gdkpixmap =
    gdk_pixmap_colormap_create_from_xpm(NULL, colormap, mask, NULL, pathname);
  if (gdkpixmap == NULL) {
    g_warning(_("Couldn't create pixmap from file: %s"), pathname);
    return;
  }

  return;
}

GList *row_select_find(GList * list, gint row)
{
  gint x;

  while (list) {
    x = GPOINTER_TO_INT(list->data);
    if (row == x)
      return list;
    list = g_list_next(list);
  }
  return NULL;
}

gboolean on_local_button_press_event(GtkWidget * widget, GdkEventButton * event,
				     gpointer user_data)
{
  if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1))
    local_dblclick();
  return FALSE;
}

void on_local_textcombo_activate(GtkEditable * editable, gpointer user_data)
{
  gchar *dir = gtk_entry_get_text(GTK_ENTRY(editable));

  chdir(dir);
  change_local();
}

void on_local_select_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			 gpointer user_data)
{
  if (!row_select_find(local_row, row)) {
    local_row = g_list_append(local_row, GINT_TO_POINTER(row));
  }
}

void on_local_drag_data_get(GtkWidget * widget, GdkDragContext * drag_context,
			    GtkSelectionData * data, guint info, guint time,
			    gpointer user_data)
{
  char cwd[SMALL_BUFFER], mydata[BIG_BUFFER];
  GList *r;
  gint src;
  struct afile *f;

  *mydata = '\0';
  for (r = local_row; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.local), src);
    g_return_if_fail(f != NULL);

    getcwd(cwd, SMALL_BUFFER);
    g_snprintf(mydata, BIG_BUFFER, "%sfile:%s/%s\n", mydata, cwd, f->file);
  }
  gtk_selection_data_set(data, data->target, 8, mydata, strlen(mydata) + 1);
}

void on_local_up_clicked(GtkButton * button, gpointer user_data)
{
  chdir("..");
  change_local();
}

void on_local_unselect_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			   gpointer user_data)
{
  GList *x = row_select_find(local_row, row);

  if (!x)
    return;

  local_row = g_list_remove(local_row, x->data);
}

void on_local_unselect_all(GtkCList * clist, gpointer user_data)
{
  g_print("unselect all for local\n");
}

void on_dir_ok_clicked(GtkButton * button, gpointer user_data)
{
  char *dir;
  int local, status;

  dir = gtk_entry_get_text(GTK_ENTRY(deadftp.newDir.directory_name));
  local =
    GPOINTER_TO_INT(gtk_object_get_data
		    (GTK_OBJECT(deadftp.newDir.dialog), "isLocal"));

  if (dir && local) {
    status =
      mkdir(dir,
	    S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
    if (status < 0) {
      char *error = g_strerror(errno), *message;
      GtkWidget *mbox;

      message = g_strconcat(_("Couldn't make directory: "), error, NULL);
      mbox =
	gnome_message_box_new(message, GNOME_MESSAGE_BOX_ERROR,
			      GNOME_STOCK_BUTTON_OK, NULL);
      gtk_widget_show(mbox);
      g_free(message);
    } else
      change_local();
  } else if (dir && !local) {
    /*ftp_send_command(FTP_STATE_MKDIR, FTP_CMD_FLAG_LIST, _("Creating Directory"), "MKD %s", dir); */
  }
  gtk_widget_destroy(deadftp.newDir.dialog);
  deadftp.newDir.dialog = NULL;
}

void on_dir_cancel_clicked(GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy(deadftp.newDir.dialog);
  deadftp.newDir.dialog = NULL;
}

void init_local_dir_layout()
{
  int i;
  GtkWidget *box, *label;

  gtk_clist_freeze(GTK_CLIST(deadftp.local));

  for (i = 1; i < NUM_DIR_LAYOUTS; i++) {
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.local), i, FALSE);
  }

  for (i = 0; i < NUM_DIR_LAYOUTS; i++) {
    label = gtk_label_new("");
    box = gtk_hbox_new(FALSE, 0);

    gtk_widget_show(label);
    gtk_widget_show(box);

    gtk_box_pack_start_defaults(GTK_BOX(box), label);
    gtk_clist_set_column_widget(GTK_CLIST(deadftp.local), i, box);

    deadftp.local_boxs[i] = box;
    deadftp.local_titles[i] = label;
  }

  if (deadftp_prefs.directory_layout[DIR_LAYOUT_FILE] != -1) {
    label = deadftp.local_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_FILE]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_FILE]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.local),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_FILE],
				    TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH] != -1) {
    label = deadftp.local_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_LENGTH]));

    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.local),
				    deadftp_prefs.
				    directory_layout[DIR_LAYOUT_LENGTH], TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_MODE] != -1) {
    label = deadftp.local_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_MODE]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_MODE]));

    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.local),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_MODE],
				    TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER] != -1) {
    label = deadftp.local_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_OWNER]));

    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.local),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER],
				    TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP] != -1) {
    label = deadftp.local_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_GROUP]));

    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.local),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP],
				    TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_DATE] != -1) {
    label = deadftp.local_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_DATE]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_DATE]));

    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.local),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_DATE],
				    TRUE);
  }

  for (i = 0; i < NUM_DIR_LAYOUTS; i++) {
    gtk_clist_set_column_auto_resize(GTK_CLIST(deadftp.local), i,
				     deadftp_prefs.local_auto_size);
  }

  if (deadftp_prefs.directory_layout[deadftp_prefs.local_sort_field] == -1) {
    for (deadftp_prefs.local_sort_field = 0;
	 (deadftp_prefs.directory_layout[deadftp_prefs.local_sort_field] == -1)
	 && deadftp_prefs.local_sort_field < NUM_DIR_LAYOUTS;
	 deadftp_prefs.local_sort_field++) {
      ;				/* do nothing */
    }
  }
  box =
    deadftp.local_boxs[deadftp_prefs.
		       directory_layout[deadftp_prefs.local_sort_field]];
  deadftp.local_arrow =
    load_pixmap(deadftp.window,
		(deadftp_prefs.local_sort_order ==
		 DIR_SORT_ASC) ? "asc.xpm" : "dsc.xpm", 0);
  gtk_box_pack_end(GTK_BOX(box), deadftp.local_arrow, FALSE, FALSE, 1);
  gtk_widget_show(deadftp.local_arrow);

  gtk_clist_thaw(GTK_CLIST(deadftp.local));
}

void load_rename_dialog()
{
  GladeXML *xml;

  /* only allow one to be opened at a time */
  g_return_if_fail(deadftp.renameDialog.dialog != NULL);

  xml = glade_xml_new(deadftp.gladefile, "renameDialog");
  deadftp.renameDialog.dialog = glade_xml_get_widget(xml, "renameDialog");

  STORE_WIDGET(deadftp.renameDialog, filename);
  STORE_WIDGET(deadftp.renameDialog, newfilename);

  glade_xml_signal_autoconnect(xml);
}

void on_rename_ok_clicked(GtkButton * button, gpointer user_data)
{
  char *filename, *newfile;
  extern void remote_do_rename(const char *filename, const char *newfile);

  newfile = gtk_entry_get_text(GTK_ENTRY(deadftp.renameDialog.newfilename));
  filename =
    gtk_object_get_data(GTK_OBJECT(deadftp.renameDialog.dialog), "local_filename");
  if (filename) {
    local_do_rename(filename, newfile);
  } else {
    filename =
      gtk_object_get_data(GTK_OBJECT(deadftp.renameDialog.dialog),
			  "remote_filename");

    if (filename) {
      remote_do_rename(filename, newfile);
    } else {
      g_assert_not_reached();
    }
  }
  gtk_widget_destroy(deadftp.renameDialog.dialog);
}

void on_rename_cancel_clicked(GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy(deadftp.renameDialog.dialog);
}

void local_do_rename(const char *oldfile, const char *newfile)
{
  g_return_if_fail(oldfile != NULL);
  g_return_if_fail(newfile != NULL);

  rename(oldfile, newfile);
}

void on_local_refresh()
{
  change_local();
}

GHashTable *pixbuf_hash = NULL;

char *get_file_pixmap(const char *filename, gboolean is_dir)
{
  char *foundfile = NULL;

  g_return_val_if_fail(filename != NULL, NULL);

  if (is_dir) {
    foundfile = find_pixmap_file("dir.xpm");
  } else {
    /* try to determine the mime type of the file using gnome functions */
    const char *mimetype = gnome_mime_type(filename);
    if (mimetype) {
      const char *pixfile = gnome_mime_get_value(mimetype, "icon-filename");
      if (pixfile)
	foundfile = find_pixmap_file(pixfile);
      else
	foundfile = NULL;
    }
    if (!mimetype || !foundfile)
      foundfile = find_pixmap_file("file.xpm");
  }
  return g_strdup(foundfile);
}

/* Attempts to determine file type by suffix and sets the icon appropriately.  */
void set_clist_icon(GtkWidget * clist, int row, int column, char *name, int is_dir)
{
  GdkPixbuf *pb = NULL, *pb_scale;
  GdkPixmap *pm = NULL;
  GdkBitmap *bm = NULL;
  char *foundfile = NULL;
  int width, height;

  g_return_if_fail(clist != NULL);
  g_return_if_fail(column >= 0);
  g_return_if_fail(name != NULL);

  foundfile = get_file_pixmap(name, is_dir);

  if (!pixbuf_hash)
    pixbuf_hash = g_hash_table_new(g_str_hash, g_str_equal);

  if (!(pb_scale = (GdkPixbuf *) g_hash_table_lookup(pixbuf_hash, foundfile))) {
    pb = gdk_pixbuf_new_from_file(foundfile);

    if (!pb)
      return;

    width = gdk_pixbuf_get_width(pb);
    height = gdk_pixbuf_get_height(pb);

    /* calculate aspect ratio and use that to determine the width from the clist row_height */
    pb_scale =
      gdk_pixbuf_scale_simple(pb, (width / height) * GTK_CLIST(clist)->row_height,
			      GTK_CLIST(clist)->row_height, GDK_INTERP_TILES);

    g_hash_table_insert(pixbuf_hash, foundfile, pb_scale);

    gdk_pixbuf_unref(pb);
  }

  /* render the scaled pixbuf into the GdkPixmap and GdkBitmap */
  gdk_pixbuf_render_pixmap_and_mask(pb_scale, &pm, &bm, 128);

  gtk_clist_set_pixtext(GTK_CLIST(clist), row, column, name, 3, pm, bm);

  gdk_pixmap_unref(pm);
  gdk_bitmap_unref(bm);
  g_free(foundfile);
}

char *find_pixmap_file(const char *pixfile)
{
  gchar *dirname, *file;

  if (g_path_is_absolute(pixfile) && g_file_exists(pixfile))
    return g_strdup(pixfile);

  dirname = g_dirname(deadftp.gladefile);
  file = g_strdup_printf("%s/%s", dirname, pixfile);
  g_free(dirname);

  if (!g_file_exists(file)) {
    g_warning("Couldn't find image file: %s\n", pixfile);
    g_free(file);
    return NULL;
  }
  return file;
}

void on_local_click_column(GtkCList * clist, gint column)
{
  int field = 0;

  while ((field < NUM_DIR_LAYOUTS)
	 && (deadftp_prefs.directory_layout[field] != column))
    field++;

  if (field == NUM_DIR_LAYOUTS) {
    /* Should never happen */
    return;
  }
  if (field == deadftp_prefs.local_sort_field) {
    local_set_sort_info(deadftp_prefs.local_sort_field,
			!deadftp_prefs.local_sort_order);
  } else {
    local_set_sort_info(field, DIR_SORT_ASC);
  }
}

static void local_set_sort_info(gint field, gboolean order)
{
  GtkWidget *box;

  /* If there is no change, do nothing */
  if ((deadftp_prefs.local_sort_field == field)
      && (deadftp_prefs.local_sort_order == order))
    return;

  box =
    deadftp.local_boxs[deadftp_prefs.
		       directory_layout[deadftp_prefs.local_sort_field]];
  gtk_container_remove(GTK_CONTAINER(box), deadftp.local_arrow);

  deadftp_prefs.local_sort_field = field;
  deadftp_prefs.local_sort_order = order;

  local_files =
    g_list_sort(local_files, dir_sort_funcs[deadftp_prefs.local_sort_field].func);
  /* XXX: should destroy old widget shouldn't we? */
  deadftp.local_arrow =
    load_pixmap(deadftp.window, (order == DIR_SORT_ASC) ? "asc.xpm" : "dsc.xpm", 0);
  gtk_widget_show(deadftp.local_arrow);

  box =
    deadftp.local_boxs[deadftp_prefs.
		       directory_layout[deadftp_prefs.local_sort_field]];
  gtk_box_pack_end(GTK_BOX(box), deadftp.local_arrow, FALSE, FALSE, 1);

  local_refresh_dir();
}

void on_local_arrange_name()
{
  /* Only sort by a field that is visible */
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_FILE] != -1)
    local_set_sort_info(DIR_LAYOUT_FILE, deadftp_prefs.local_sort_order);
}

void on_local_arrange_owner()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER] != -1)
    local_set_sort_info(DIR_LAYOUT_OWNER, deadftp_prefs.local_sort_order);
}

void on_local_arrange_group()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP] != -1)
    local_set_sort_info(DIR_LAYOUT_GROUP, deadftp_prefs.local_sort_order);
}

void on_local_arrange_mode()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_MODE] != -1)
    local_set_sort_info(DIR_LAYOUT_MODE, deadftp_prefs.local_sort_order);
}

void on_local_arrange_date()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_DATE] != -1)
    local_set_sort_info(DIR_LAYOUT_DATE, deadftp_prefs.local_sort_order);
}

void on_local_arrange_length()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH] != -1)
    local_set_sort_info(DIR_LAYOUT_LENGTH, deadftp_prefs.local_sort_order);
}

void on_local_arrange_ascending()
{
  local_set_sort_info(deadftp_prefs.local_sort_field, DIR_SORT_ASC);
}

void on_local_arrange_descending()
{
  local_set_sort_info(deadftp_prefs.local_sort_field, DIR_SORT_DSC);
}

void on_local_arrange_autosize(GtkMenuItem * item)
{
  int i;
  gboolean active = GTK_CHECK_MENU_ITEM(item)->active;

  /* If there's no change, do nothing */
  if (deadftp_prefs.local_auto_size == active)
    return;

  deadftp_prefs.local_auto_size = active;
  for (i = 0; i < NUM_DIR_LAYOUTS; i++)
    gtk_clist_set_column_auto_resize(GTK_CLIST(deadftp.local), i,
				     deadftp_prefs.local_auto_size);
  local_refresh_dir();
}

void on_local_move()
{
  GtkWidget *dialog =
    gnome_request_dialog(FALSE, _("Move file(s) where? "), "", REGULAR_BUFFER,
			 local_move_string_callback, (gpointer) 0,
			 GTK_WINDOW(deadftp.window));
  gtk_widget_show(dialog);
}

static void local_move_string_callback(gchar * string, gpointer data)
{
  GList *l;

  if (!string)
    return;

  /* The dialog was modal so the selected files should still be the same */
  for (l = local_row; l; l = g_list_next(l)) {
    gint src = GPOINTER_TO_INT(l->data);
    struct afile *f =
      (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.local), src);
    gchar *str;

    g_return_if_fail(f != NULL);

    str = g_strdup_printf("%s/%s", string, f->file);
    /* XXX: do error checking */
    rename(f->file, str);
    g_free(str);
  }
  gtk_clist_unselect_all(GTK_CLIST(deadftp.local));
  g_list_free(local_row);
  local_row = NULL;
  change_local();
}

void on_local_mode()
{
  load_mode_dialog(TRUE);
}

void load_mode_dialog(gboolean local)
{
  GladeXML *xml;
  gchar *str;
  struct afile *file;

  if ((local && !local_row) || (!local && !remote_row))
    return;

  xml = glade_xml_new(deadftp.gladefile, "modeDialog");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  STORE_WIDGET(deadftp.modeDialog, mode);
  STORE_WIDGET(deadftp.modeDialog, mode_user_read);
  STORE_WIDGET(deadftp.modeDialog, mode_user_write);
  STORE_WIDGET(deadftp.modeDialog, mode_user_exe);
  STORE_WIDGET(deadftp.modeDialog, mode_group_read);
  STORE_WIDGET(deadftp.modeDialog, mode_group_write);
  STORE_WIDGET(deadftp.modeDialog, mode_group_exe);
  STORE_WIDGET(deadftp.modeDialog, mode_other_read);
  STORE_WIDGET(deadftp.modeDialog, mode_other_write);
  STORE_WIDGET(deadftp.modeDialog, mode_other_exe);

  glade_xml_signal_autoconnect(xml);
  gtk_object_unref(GTK_OBJECT(xml));

  deadftp.modeDialog.dont_handle_signal = FALSE;
  gtk_object_set_data(GTK_OBJECT(deadftp.modeDialog.mode), "isLocal",
		      GINT_TO_POINTER(local));
  file =
    (struct afile *) g_list_nth_data(local ? local_files : remote_files,
				     GPOINTER_TO_INT(local ? local_row->
						     data : remote_row->data));
  str = g_strdup_printf("%o", str_to_mode(file->mode));
  /* The changed signal handler for the entry will take care of setting the checkbuttons */
  gtk_entry_set_text(GTK_ENTRY(deadftp.modeDialog.mode), str);
  g_free(str);
}

void on_mode_toggled(GtkToggleButton * button, gpointer data)
{
  mode_t mode =
    strtol(gtk_entry_get_text(GTK_ENTRY(deadftp.modeDialog.mode)), NULL, 8);
  gchar mode_str[8];

  if (deadftp.modeDialog.dont_handle_signal)
    return;
  deadftp.modeDialog.dont_handle_signal = TRUE;

  switch (atoi((char *) data)) {
    /* User Read */
  case 0:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IRUSR;
    else
      mode &= ~S_IRUSR;
    break;
    /* User Write */
  case 1:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IWUSR;
    else
      mode &= ~S_IWUSR;
    break;
    /* User Execute */
  case 2:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IXUSR;
    else
      mode &= ~S_IXUSR;
    break;
    /* Group Read */
  case 3:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IRGRP;
    else
      mode &= ~S_IRGRP;
    break;
    /* Group Write */
  case 4:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IWGRP;
    else
      mode &= ~S_IWGRP;
    break;
    /* Group Execute */
  case 5:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IXGRP;
    else
      mode &= ~S_IXGRP;
    break;
    /* Other Read */
  case 6:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IROTH;
    else
      mode &= ~S_IROTH;
    break;
    /* Other Write */
  case 7:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IWOTH;
    else
      mode &= ~S_IWOTH;
    break;
    /* Other Execute */
  case 8:
    if (gtk_toggle_button_get_active(button))
      mode |= S_IXOTH;
    else
      mode &= ~S_IXOTH;
    break;
  default:
    g_assert_not_reached();
  }
  g_snprintf(mode_str, 8, "%o", mode);
  gtk_entry_set_text(GTK_ENTRY(deadftp.modeDialog.mode), mode_str);

  deadftp.modeDialog.dont_handle_signal = FALSE;
}

void on_mode_changed(GtkEntry * entry, gpointer data)
{
  mode_t mode = strtol(gtk_entry_get_text(GTK_ENTRY(entry)), NULL, 8);
  gchar mode_str[8];

  if (deadftp.modeDialog.dont_handle_signal)
    return;
  deadftp.modeDialog.dont_handle_signal = TRUE;

  mode &= ALLPERMS;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(deadftp.modeDialog.mode_user_read),
			       mode & S_IRUSR);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(deadftp.modeDialog.mode_user_write),
			       mode & S_IWUSR);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(deadftp.modeDialog.mode_user_exe),
			       mode & S_IXUSR);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(deadftp.modeDialog.mode_group_read),
			       mode & S_IRGRP);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON
			       (deadftp.modeDialog.mode_group_write),
			       mode & S_IWGRP);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(deadftp.modeDialog.mode_group_exe),
			       mode & S_IXGRP);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(deadftp.modeDialog.mode_other_read),
			       mode & S_IROTH);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON
			       (deadftp.modeDialog.mode_other_write),
			       mode & S_IWOTH);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(deadftp.modeDialog.mode_other_exe),
			       mode & S_IXOTH);

  g_snprintf(mode_str, 8, "%o", mode);
  gtk_entry_set_text(GTK_ENTRY(entry), mode_str);

  deadftp.modeDialog.dont_handle_signal = FALSE;
}

void on_mode_ok_clicked(GtkButton * button, gpointer data)
{
  mode_t mode =
    strtol(gtk_entry_get_text(GTK_ENTRY(deadftp.modeDialog.mode)), NULL, 8);
  GList *l;
  gboolean local;

  local =
    GPOINTER_TO_INT(gtk_object_get_data
		    (GTK_OBJECT(deadftp.modeDialog.mode), "isLocal"));
  for (l = local ? local_row : remote_row; l; l = g_list_next(l)) {
    gint src = GPOINTER_TO_INT(l->data);
    struct afile *f;

    if (local) {
      f = (struct afile *) g_list_nth_data(local_files, src);
      chmod(f->file, mode);
    } else {
      f = (struct afile *) g_list_nth_data(remote_files, src);
      /* ftp_send_command(FTP_STATE_CHMOD, FTP_CMD_FLAG_LIST, _("Changing File Attributes"), "SITE CHMOD %o %s", mode, f->file); */
    }
  }
  if (local)
    change_local();
}

void on_local_transfer()
{
  /* XXX: implement this */
}
