/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <strings.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "globals.h"
#include "socket.h"
#include "proxy.h"
#include "socks.h"

/* Local functions */
static void socks5_authenticate(socket_ctx_t * sock);
static void socks5_authenticate_1(socket_ctx_t * sock, gboolean ok);
static void socks5_authenticate_2(socket_ctx_t * sock, gboolean ok);
static void socks5_connect(socket_ctx_t * sock);
static void socks5_connect_1(socket_ctx_t * sock);
static void socks5_connect_2(socket_ctx_t * sock, gboolean ok);
static void socks5_bind(socket_ctx_t * sock);
static void socks5_bind_1(socket_ctx_t * sock);
static void socks5_bind_2(socket_ctx_t * sock, gboolean ok);
static void socks5_accept(socket_ctx_t * sock);
static void socks5_accept_1(socket_ctx_t * sock, gboolean ok);
static void socks4_connect(socket_ctx_t * sock);
static void socks4_connect_1(socket_ctx_t * sock, gboolean ok);
static void socks4_bind(socket_ctx_t * sock);
static void socks4_bind_1(socket_ctx_t * sock, gboolean ok);
static void socks4_accept(socket_ctx_t * sock);
static void socks4_accept_1(socket_ctx_t * sock, gboolean ok);
static void proxy_write(gpointer data, gint source, GdkInputCondition condition);
static void proxy_read(gpointer data, gint source, GdkInputCondition condition);

proxy_desc_t proxies[] = {
  /* 0 - NONE */
  {N_("No Proxy Server"), PROXY_FLAG_DISABLE | PROXY_FLAG_PORT, NULL, NULL, NULL},
  /* 1 - PROXY_USER1 */
  {N_("Anonymous USER (username@host port)"), 0, NULL, NULL, NULL},
  /* 2 - PROXY_USER2 */
  {N_("Anonymous USER (username@host:port)"), 0, NULL, NULL, NULL},
  /* 3 - PROXY_USER_AUTH1 */
  {N_("Authenticated USER (username@host port)"), PROXY_FLAG_AUTH,
   NULL, NULL, NULL},
  /* 4 - PROXY_USER_AUTH2 */
  {N_("Authenticated USER (username@host:port)"), PROXY_FLAG_AUTH,
   NULL, NULL, NULL},
  /* 5 - SOCKS4 */
  {N_("SOCKS 4 Proxy Server"), PROXY_FLAG_AUTH | PROXY_FLAG_PORT, socks4_connect,
   socks4_bind, socks4_accept},
  /* 6 - SOCKS5 */
  {N_("SOCKS 5 Proxy Server"), PROXY_FLAG_AUTH | PROXY_FLAG_PORT, socks5_connect,
   socks5_bind, socks5_accept},
  {NULL, 0, NULL, NULL, NULL},
};

static void proxy_send_data(socket_ctx_t * sock,
			    void (*callback) (socket_ctx_t *, gboolean))
{
  int len;

  g_return_if_fail(sock->proxy != NULL);

  sock->proxy->callback = callback;
  if ((len =
       send(sock->sockfd, sock->proxy->buffer, sock->proxy->bufferlen, 0)) <= 0) {
    if (errno == EAGAIN)
      sock->proxy->watchid =
	gdk_input_add(sock->sockfd, GDK_INPUT_WRITE, proxy_write, (gpointer) sock);
    else
      callback(sock, FALSE);
    return;
  }
  if (len == sock->proxy->bufferlen) {
    sock->proxy->watchid =
      gdk_input_add(sock->sockfd, GDK_INPUT_READ, proxy_read, (gpointer) sock);
    return;
  }
  memcpy(sock->proxy->buffer, sock->proxy->buffer + len,
	 sock->proxy->bufferlen - len);
  sock->proxy->bufferlen -= len;
  gdk_input_add(sock->sockfd, GDK_INPUT_WRITE, proxy_write, (gpointer) sock);
}

static void proxy_write(gpointer data, gint source, GdkInputCondition condition)
{
  socket_ctx_t *sock = (socket_ctx_t *) data;
  int len = 0;

  if (sock->sockfd != source)
    return;

  while (sock->proxy->bufferlen
	 && (len =
	     send(sock->sockfd, sock->proxy->buffer + len, sock->proxy->bufferlen,
		  0)))
    sock->proxy->bufferlen -= len;

  /* an error occurred, call callback to indicate it */
  if ((len < 0 && errno != EAGAIN) || len == 0) {
    sock->proxy->callback(sock, FALSE);
    return;
  }
  /* not everything was sent so wait until we can send again */
  if (sock->proxy->bufferlen) {
    memcpy(sock->proxy->buffer, sock->proxy->buffer + len, sock->proxy->bufferlen);
    return;
  }
  /* If everything was sent, remove the write input callback and add a read input callback */
  gdk_input_remove(sock->proxy->watchid);
  sock->proxy->watchid =
    gdk_input_add(sock->sockfd, GDK_INPUT_READ, proxy_read, (gpointer) sock);
}

static void proxy_read(gpointer data, gint source, GdkInputCondition condition)
{
  socket_ctx_t *sock = (socket_ctx_t *) data;

  if (sock->sockfd != source)
    return;

  /* got data on socket, remote input tag and call callback */
  gdk_input_remove(sock->proxy->watchid);
  sock->proxy->callback(sock, TRUE);
}

/* We already have a connection to the SOCKS server, now authenticate */
static void socks5_authenticate(socket_ctx_t * sock)
{
  struct socks5_version_hdr *s5v;

  g_return_if_fail(sock != NULL);
  g_return_if_fail(sock->proxy != NULL);

  s5v = (struct socks5_version_hdr *) sock->proxy->buffer;
  s5v->version = SOCKS5_VERSION;
  s5v->nmethods = 1;
  sock->proxy->bufferlen = sizeof(struct socks5_version_hdr);
  s5v->methods[0] = SOCKS5_AUTH_NONE;
  sock->proxy->bufferlen += sizeof(s5v->methods[0]);

  if (deadftp_prefs.firewall.proxy_username
      && *deadftp_prefs.firewall.proxy_username) {
    s5v->methods[1] = SOCKS5_AUTH_USER;
    sock->proxy->bufferlen += sizeof(s5v->methods[1]);
    s5v->nmethods++;
  }

  proxy_send_data(sock, socks5_authenticate_1);
}

static void socks5_authenticate_1(socket_ctx_t * sock, gboolean ok)
{
  struct socks5_version_rpl *s5r;

  if (!ok) {
    socket_error(sock);
    return;
  }

  if (recv(sock->sockfd, sock->proxy->buffer, SMALL_BUFFER, 0) < 0) {
    socket_error(sock);
    return;
  }

  s5r = (struct socks5_version_rpl *) sock->proxy->buffer;
  if (s5r->method == SOCKS5_AUTH_INVALID) {
    socket_error(sock);
    return;
  }

  if (s5r->method == SOCKS5_AUTH_USER) {
    struct socks5_user_auth_hdr *s5u;
    guint8 *b;

    sock->proxy->bufferlen = sizeof(struct socks5_user_auth_hdr);
    s5u = (struct socks5_user_auth_hdr *) sock->proxy->buffer;
    s5u->version = 1;

    b = sock->proxy->buffer + sock->proxy->bufferlen;
    if (deadftp_prefs.firewall.proxy_username) {
      s5u->user_len = strlen(deadftp_prefs.firewall.proxy_username);
      sock->proxy->bufferlen += s5u->user_len;
      strcpy(b, deadftp_prefs.firewall.proxy_username);
      b += s5u->user_len;
    } else {
      s5u->user_len = 0;
    }
    if (deadftp_prefs.firewall.proxy_password) {
      *b++ = strlen(deadftp_prefs.firewall.proxy_password);
      sock->proxy->bufferlen += 1;
      strcpy(b, deadftp_prefs.firewall.proxy_password);
      sock->proxy->bufferlen += strlen(deadftp_prefs.firewall.proxy_password);
      b += strlen(deadftp_prefs.firewall.proxy_password);
    } else {
      *b = 0;
      sock->proxy->bufferlen += 1;
    }

    proxy_send_data(sock, socks5_authenticate_2);
  }
  /* All ready to go, go back to the proper function */
  if (sock->proxy->sock_type == SOCK_CONNECT)
    socks5_connect_1(sock);
  else
    socks5_bind_1(sock);
}

static void socks5_authenticate_2(socket_ctx_t * sock, gboolean ok)
{
  struct socks5_user_auth_rpl *s5ur;

  if (!ok) {
    socket_error(sock);
    return;
  }

  if (recv(sock->sockfd, sock->proxy->buffer, sizeof(struct socks5_user_auth_rpl), 0)
      < 0) {
    socket_error(sock);
    return;
  }
  s5ur = (struct socks5_user_auth_rpl *) sock->proxy->buffer;
  if (s5ur->status != SOCKS5_REPLY_SUCCESS) {
    socket_error(sock);
    return;
  }
  /* everything is OK, call proper function to continue with connection */
  if (sock->proxy->sock_type == SOCK_CONNECT)
    socks5_connect_1(sock);
  else
    socks5_bind_1(sock);
}

static void socks5_connect(socket_ctx_t * sock)
{
  g_return_if_fail(sock->proxy != NULL);
  sock->proxy->sock_type = SOCK_CONNECT;

  socks5_authenticate(sock);
}

static void socks5_connect_1(socket_ctx_t * sock)
{
  struct socks5_request_hdr *s5r;

  s5r = (struct socks5_request_hdr *) sock->proxy->buffer;
  s5r->version = SOCKS5_VERSION;
  s5r->cmd = SOCKS5_CMD_CONNECT;
  s5r->reserved = 0;
  s5r->atype = SOCKS5_ADDR_TYPE_IP;
  s5r->dst_addr = sock->proxy->addr;
  s5r->dst_port = sock->proxy->port;
  sock->proxy->bufferlen = sizeof(struct socks5_request_hdr);

  proxy_send_data(sock, socks5_connect_2);
}

static void socks5_connect_2(socket_ctx_t * sock, gboolean ok)
{
  struct socks5_request_rpl *s5rr;

  if (!ok) {
    socket_error(sock);
    return;
  }

  if (recv(sock->sockfd, sock->proxy->buffer, SMALL_BUFFER, 0) < 0) {
    socket_error(sock);
    return;
  }

  s5rr = (struct socks5_request_rpl *) sock->proxy->buffer;
  if (s5rr->reply != SOCKS5_REPLY_SUCCESS) {
    socket_error(sock);
    return;
  }
  /* since the SOCKS server will probably return 0 for the bnd_addr when
   * we do a bind, store the value now */
  /* sock->proxy->serv_addr = s5rr->bnd_addr; */

  /* Connected!  finish socket */
  socket_finish(sock);
}

static void socks5_bind(socket_ctx_t * sock)
{
  g_return_if_fail(sock->proxy != NULL);
  sock->proxy->sock_type = SOCK_BIND;

  socks5_authenticate(sock);
}

static void socks5_bind_1(socket_ctx_t * sock)
{
  struct socks5_request_hdr *s5r;

  s5r = (struct socks5_request_hdr *) sock->proxy->buffer;
  s5r->version = SOCKS5_VERSION;
  s5r->cmd = SOCKS5_CMD_BIND;
  s5r->reserved = 0;
  s5r->atype = SOCKS5_ADDR_TYPE_IP;
  /* s5r->dst_addr = socks.addr; */
  /* s5r->dst_port = socks.port; */
  sock->proxy->bufferlen = sizeof(struct socks5_request_hdr);

  proxy_send_data(sock, socks5_bind_2);
}

static void socks5_bind_2(socket_ctx_t * sock, gboolean ok)
{
  struct socks5_request_rpl *s5rr;

  if (!ok) {
    socket_error(sock);
    return;
  }

  if (recv(sock->sockfd, sock->proxy->buffer, sizeof(struct socks5_request_rpl), 0) <
      0) {
    socket_error(sock);
    return;
  }

  s5rr = (struct socks5_request_rpl *) sock->proxy->buffer;
  if (s5rr->reply != SOCKS5_REPLY_SUCCESS) {
    socket_error(sock);
    return;
  }

  if (s5rr->bnd_addr) {
    /* If the server returns something other that 0 (I have yet to see this happen) use
     * that instead of the value we got while we connected.  */
    sock->addr.sin_addr.s_addr = s5rr->bnd_addr;
  }
  sock->addr.sin_port = s5rr->bnd_port;

  g_warning("SOCKS5: bound to socket on port %d %d\n", s5rr->bnd_port,
	    ntohs(s5rr->bnd_port));
  /* Everything's ready to go, finish up */
  socket_finish(sock);
}

static void socks5_accept(socket_ctx_t * sock)
{
  sock->proxy->callback = socks5_accept_1;
  sock->proxy->watchid =
    gdk_input_add(sock->sockfd, GDK_INPUT_READ, proxy_read, (gpointer) sock);
}

static void socks5_accept_1(socket_ctx_t * sock, gboolean ok)
{
  struct socks5_request_rpl *s5rr;

  g_return_if_fail(sock != NULL);

  if (recv(sock->sockfd, sock->proxy->buffer, sizeof(struct socks5_request_rpl), 0) <
      0) {
    socket_error(sock);
    return;
  }

  s5rr = (struct socks5_request_rpl *) sock->proxy->buffer;
  if (s5rr->reply != SOCKS5_REPLY_SUCCESS) {
    socket_error(sock);
    return;
  }
  sock->addr.sin_addr.s_addr = s5rr->bnd_addr;
  sock->addr.sin_port = s5rr->bnd_port;

  socket_finish(sock);
}

static void socks4_connect(socket_ctx_t * sock)
{
  struct socks4_request_hdr *s4r;

  s4r = (struct socks4_request_hdr *) sock->proxy->buffer;
  s4r->version = SOCKS4_VERSION;
  s4r->command = SOCKS4_CMD_CONNECT;
  s4r->dst_addr = sock->proxy->addr;
  s4r->dst_port = sock->proxy->port;
  strncpy(s4r->userid, deadftp_prefs.firewall.proxy_username,
	  SMALL_BUFFER - sizeof(struct socks5_request_hdr));
  sock->proxy->bufferlen =
    sizeof(struct socks5_request_hdr) +
    strlen(deadftp_prefs.firewall.proxy_username) + 1;

  proxy_send_data(sock, socks4_connect_1);
}

static void socks4_connect_1(socket_ctx_t * sock, gboolean ok)
{
  struct socks4_reply_hdr *s4r;

  if (!ok) {
    socket_error(sock);
    return;
  }

  if (recv(sock->sockfd, sock->proxy->buffer, sizeof(struct socks4_reply_hdr), 0) <
      0) {
    socket_error(sock);
    return;
  }

  s4r = (struct socks4_reply_hdr *) sock->proxy->buffer;
  if (s4r->command != SOCKS4_REPLY_SUCCESS) {
    socket_error(sock);
    return;
  }

  sock->addr.sin_addr.s_addr = s4r->dst_addr;
  sock->addr.sin_port = s4r->dst_port;

  /* Everything's ready to go, finish up */
  socket_finish(sock);
}

static void socks4_bind(socket_ctx_t * sock)
{
  struct socks4_request_hdr *s4r;

  s4r = (struct socks4_request_hdr *) sock->proxy->buffer;
  s4r->version = SOCKS4_VERSION;
  s4r->command = SOCKS4_CMD_BIND;
  /* s4r->dst_addr = sock->addr; */
  /* s4r->dst_port = sock->port; */
  strncpy(s4r->userid, deadftp_prefs.firewall.proxy_username,
	  SMALL_BUFFER - sizeof(struct socks4_request_hdr) - 1);
  sock->proxy->bufferlen =
    sizeof(struct socks4_request_hdr) +
    strlen(deadftp_prefs.firewall.proxy_username) + 1;

  proxy_send_data(sock, socks4_bind_1);
}

static void socks4_bind_1(socket_ctx_t * sock, gboolean ok)
{
  struct socks4_reply_hdr *s4r;

  if (!ok) {
    socket_error(sock);
    return;
  }

  if (recv(sock->sockfd, sock->proxy->buffer, sizeof(struct socks4_reply_hdr), 0) <
      0) {
    socket_error(sock);
    return;
  }

  s4r = (struct socks4_reply_hdr *) sock->proxy->buffer;
  if (s4r->command != SOCKS4_REPLY_SUCCESS) {
    socket_error(sock);
    return;
  }

  if (s4r->dst_addr) {
    sock->addr.sin_addr.s_addr = s4r->dst_addr;
  }
  sock->addr.sin_port = s4r->dst_port;

  g_warning("SOCKS5: bound to socket on port %d %d\n", s4r->dst_port,
	    ntohs(s4r->dst_port));
  /* Everything's ready to go, finish up */
  socket_finish(sock);
}

static void socks4_accept(socket_ctx_t * sock)
{
  sock->proxy->callback = socks4_accept_1;
  sock->proxy->watchid =
    gdk_input_add(sock->sockfd, GDK_INPUT_READ, proxy_read, (gpointer) sock);
}

static void socks4_accept_1(socket_ctx_t * sock, gboolean ok)
{
  struct socks4_reply_hdr *s4r;

  g_return_if_fail(sock != NULL);

  if (recv(sock->sockfd, sock->proxy->buffer, sizeof(struct socks4_reply_hdr), 0) <
      0) {
    socket_error(sock);
    return;
  }

  s4r = (struct socks4_reply_hdr *) sock->proxy->buffer;
  if (s4r->command != SOCKS4_REPLY_SUCCESS) {
    socket_error(sock);
    return;
  }
  sock->addr.sin_addr.s_addr = s4r->dst_addr;
  sock->addr.sin_port = s4r->dst_port;

  socket_finish(sock);
}
