/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <strings.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <gnome.h>

#include "globals.h"
#include "main.h"
#include "socket.h"
#include "ftp.h"
#include "proxy.h"

/* Local functions */
static void *socket_connect_thread(void *arg);
static void *socket_addr_connect_thread(void *arg);
static void socket_async_connect_handler(gpointer data, gint source,
					 GdkInputCondition cond);
static void socket_async_accept_handler(gpointer data, gint source,
					GdkInputCondition cond);
void socket_lookup(const char *hostname, struct in_addr *addr);

void socket_lookup(const char *hostname, struct in_addr *addr)
{
  struct hostent *hp;

  if ((addr->s_addr = inet_addr(hostname)) != -1)
    return;

  pthread_testcancel();
  hp = gethostbyname(hostname);
  pthread_testcancel();

  if (!hp)
    return;
  memcpy((char *) addr, hp->h_addr, sizeof(struct in_addr));
}

static int socket_do_connect(socket_ctx_t * sock)
{
  pthread_testcancel();
  sock->sockfd = socket(AF_INET, SOCK_STREAM, 0);
  pthread_testcancel();

  if (sock->sockfd == -1) {
    g_error("Error creating socket: %s\n", g_strerror(errno));
    return -1;
  }

  if (socket_set_nonblock(sock, 1) == -1) {
    g_error("Error setting nonblock: %s\n", g_strerror(errno));
    return -1;
  }

  if (deadftp_prefs.firewall.bind_ip) {
    struct sockaddr_in bind_addr;
    memset(&bind_addr, 0, sizeof(bind_addr));
    bind_addr.sin_family = AF_INET;
    bind_addr.sin_port   = htons(0);
    bind_addr.sin_addr.s_addr = inet_addr(deadftp_prefs.firewall.bind_ip_address);
    pthread_testcancel();
    if (bind(sock->sockfd, (struct sockaddr *) &bind_addr, sizeof(bind_addr)) < 0) {
      g_error("Could not bind to IP address: %s\n", deadftp_prefs.firewall.bind_ip_address);
      return -1;
    }
  }

  pthread_testcancel();
  if (connect(sock->sockfd, (struct sockaddr *) &sock->addr, sizeof(sock->addr)) < 0) {
    pthread_testcancel();
    if (errno != EINPROGRESS)
      return -1;
  }
  GDK_THREADS_ENTER();
  sock->watchid =
    gdk_input_add(sock->sockfd,
		  GDK_INPUT_READ | GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION,
		  socket_async_connect_handler, sock);
  GDK_THREADS_LEAVE();

  return 0;
}

static void *socket_connect_thread(void *arg)
{
  socket_ctx_t *sock;

  sock = (socket_ctx_t *) arg;

  memset(&sock->addr, 0, sizeof(struct sockaddr_in));
  socket_lookup(sock->hostname, &sock->addr.sin_addr);
  sock->addr.sin_port = htons(sock->port);
  sock->addr.sin_family = AF_INET;

  if (deadftp_prefs.firewall.use_proxy) {
    sock->proxy = g_new0(proxy_info_t, 1);
    sock->proxy->addr = sock->addr.sin_addr.s_addr;
    sock->proxy->port = htons(sock->port);
    sock->proxy->type = deadftp_prefs.firewall.proxy_type;
    sock->proxy->sock_type = SOCK_CONNECT;

    socket_lookup(deadftp_prefs.firewall.proxy_hostname, &sock->addr.sin_addr);
    sock->addr.sin_port = htons(deadftp_prefs.firewall.proxy_port);
  }

  if (socket_do_connect(sock) < 0) {
    /* Call callback but indicate an error occurred */
    return GINT_TO_POINTER(-1);
  }

  return GINT_TO_POINTER(0);
}

static void *socket_addr_connect_thread(void *arg)
{
  socket_ctx_t *sock;

  sock = (socket_ctx_t *) arg;

  sock->addr.sin_family = AF_INET;
  sock->addr.sin_port = htons(sock->port);

  if (deadftp_prefs.firewall.use_proxy) {
    sock->proxy = g_new0(proxy_info_t, 1);
    sock->proxy->addr = sock->addr.sin_addr.s_addr;
    sock->proxy->port = htons(sock->port);
    sock->proxy->type = deadftp_prefs.firewall.proxy_type;
    sock->proxy->sock_type = SOCK_CONNECT;

    socket_lookup(deadftp_prefs.firewall.proxy_hostname, &sock->addr.sin_addr);
    sock->addr.sin_port = htons(deadftp_prefs.firewall.proxy_port);
  }

  if (socket_do_connect(sock) < 0) {
    /* Call callback indicating error */
    return GINT_TO_POINTER(-1);
  }

  return GINT_TO_POINTER(0);
}

void socket_connect(socket_ctx_t * sock, const char *hostname, int port,
		    socket_callback_func callback, gpointer data)
{
  pthread_attr_t attr;

  g_return_if_fail(hostname != NULL);

  if (sock->hostname)
    g_free(sock->hostname);
  sock->hostname = g_strdup(hostname);
  sock->port = port;
  sock->callback = callback;
  sock->data = data;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, TRUE);
  pthread_create(&sock->thread, &attr, socket_connect_thread, (void *) sock);
}

void socket_addr_connect(socket_ctx_t * sock, unsigned long addr, int port,
			 socket_callback_func callback, gpointer data)
{
  pthread_attr_t attr;

  sock->addr.sin_addr.s_addr = htonl(addr);
  sock->port = port;
  sock->callback = callback;
  sock->data = data;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, TRUE);
  pthread_create(&sock->thread, &attr, socket_addr_connect_thread, (void *) sock);
}

/* Need to rewrite this since we could hang while attempting to connect to socks server */
void socket_new_server(socket_ctx_t * sock, int port, socket_callback_func callback,
		       gpointer data)
{
  sock->callback = callback;
  sock->data = data;
  if (deadftp_prefs.firewall.bind_ip)
    sock->addr.sin_addr.s_addr = inet_addr(deadftp_prefs.firewall.bind_ip_address);
  else
    sock->addr.sin_addr.s_addr = htonl(INADDR_ANY);
  sock->addr.sin_family = AF_INET;
  if (port == 0 && deadftp_prefs.firewall.limit_port_range)
    sock->addr.sin_port = htons(deadftp_prefs.firewall.limit_port_low);
  else
    sock->addr.sin_port = htons(port);

  if (deadftp_prefs.firewall.use_proxy) {
    if (PROXY_FLAG(deadftp_prefs.firewall.proxy_type, PROXY_FLAG_PORT)) {
      callback(NULL, data);
      return;
    }

    sock->proxy = g_new0(proxy_info_t, 1);
    sock->proxy->type = deadftp_prefs.firewall.proxy_type;
    sock->proxy->sock_type = SOCK_BIND;
    sock->proxy->port = htons(port);

    if (sock->hostname)
      g_free(sock->hostname);
    sock->hostname = g_strdup(deadftp_prefs.firewall.proxy_hostname);
    sock->port = deadftp_prefs.firewall.proxy_port;

    if (socket_do_connect(sock) < 0) {
      /* Call callback indicating error */
      socket_error(sock);
      return;
    }
  } else {
    sock->sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (port == 0 && deadftp_prefs.firewall.limit_port_range) {
      int result;
      port = deadftp_prefs.firewall.limit_port_low;
      while (((result = bind(sock->sockfd, (struct sockaddr *) &sock->addr, sizeof(sock->addr))) < 0)
	     && (errno == EADDRINUSE)
	     && (port <= deadftp_prefs.firewall.limit_port_high)) {
	sock->addr.sin_port = htons(port++);
      }
      if (result < 0) {
	socket_error(sock);
	return;
      }
    } else {
      if (bind(sock->sockfd, (struct sockaddr *) &sock->addr, sizeof(sock->addr)) < 0) {
        socket_error(sock);
        return;
      }
    }
    listen(sock->sockfd, 10);

    socket_set_nonblock(sock, TRUE);
    callback(sock, sock->data);
  }
}

char *socket_make_string(socket_ctx_t * sock)
{
  struct sockaddr_in serv;
  int len;
  unsigned long addr = 0, port = 0;
  char str[SMALL_BUFFER];

  if (deadftp_prefs.firewall.use_proxy) {
    port = ntohs(sock->proxy->port);

    if (!sock->proxy->addr) {
      /* If the bnd_addr returned by the socks server is 0, use the SOCKS server's IP */
      addr = ntohl(sock->addr.sin_addr.s_addr);
    } else {
      addr = ntohl(sock->proxy->addr);
    }
  } else {
    /* getsockname on a server socket just returns 0.0.0.0 for the IP
     * address, so we need to try and get our IP address from the control
     * connection sockfd.  */
    len = sizeof(serv);
    if (getsockname(FTP.sock->sockfd, (struct sockaddr *) &serv, &len) < 0) {
      return NULL;
    }
    addr = ntohl(serv.sin_addr.s_addr);

    len = sizeof(serv);
    if (getsockname(sock->sockfd, (struct sockaddr *) &serv, &len) < 0) {
      return NULL;
    }
    port = ntohs(serv.sin_port);
  }

  g_snprintf(str, SMALL_BUFFER, "%ld,%ld,%ld,%ld,%ld,%ld", (addr >> 24) & 0xff,
	     (addr >> 16) & 0xff, (addr >> 8) & 0xff, addr & 0xff,
	     (port >> 8) & 0xff, port & 0xff);
  return g_strdup(str);
}

void socket_accept_connection(socket_ctx_t * sock, socket_callback_func callback,
			      gpointer data)
{
  sock->callback = callback;
  sock->data = data;

  if (deadftp_prefs.firewall.use_proxy) {
    if (proxies[sock->proxy->type].accept)
      proxies[sock->proxy->type].accept(sock);
    else
      socket_error(sock);
    return;
  } else {
    int newfd, len;

    len = sizeof(sock->addr);
    newfd = accept(sock->sockfd, (struct sockaddr *) &sock->addr, &len);
    if (newfd < 0 && errno == EAGAIN) {
      sock->watchid =
	gdk_input_add(sock->sockfd, GDK_INPUT_READ, socket_async_accept_handler,
		      (gpointer) sock);
      return;
    } else {
      socket_error(sock);
      return;
    }
    sock->callback(sock, sock->data);
  }
}

static void socket_async_accept_handler(gpointer data, gint source,
					GdkInputCondition cond)
{
  socket_ctx_t *sock;
  int newfd, len;

  sock = (socket_ctx_t *) data;

  if (source != sock->sockfd)
    return;

  len = sizeof(sock->addr);
  newfd = accept(sock->sockfd, (struct sockaddr *) &sock->addr, &len);
  if (newfd < 0 && errno == EAGAIN) {
    /* would have blocked for some reason so wait for another callback 
     * on the socket */
    return;
  }

  if (sock->watchid != -1)
    gdk_input_remove(sock->watchid);

  close(sock->sockfd);
  sock->sockfd = newfd;

  sock->callback(sock, sock->data);
}

static void socket_async_connect_handler(gpointer data, gint source,
					 GdkInputCondition cond)
{
  socket_ctx_t *sock;
  int error;
  socklen_t len;

  sock = (socket_ctx_t *) data;

  if (source != sock->sockfd)
    return;

  if (sock->watchid != -1)
    gdk_input_remove(sock->watchid);

  if ((cond & GDK_INPUT_READ) || (cond & GDK_INPUT_WRITE)) {
    len = sizeof(error);
    if (getsockopt(sock->sockfd, SOL_SOCKET, SO_ERROR, &error, &len) >= 0) {
      if (!error) {
	if (sock->proxy) {
	  switch (sock->proxy->sock_type) {
	  case SOCK_CONNECT:
	    if (proxies[sock->proxy->type].connect)
	      proxies[sock->proxy->type].connect(sock);
	    else
	      socket_error(sock);
	    break;
	  case SOCK_BIND:
	    if (proxies[sock->proxy->type].bind)
	      proxies[sock->proxy->type].bind(sock);
	    else
	      socket_error(sock);
	    break;
	  }
	  return;
	}

	sock->callback(sock, sock->data);

	return;
      }
    }
  }
  print_console(CONSOLE_ERROR, _("Error Connecting to host %s: %s\n"),
		sock->hostname ? sock->hostname : inet_ntoa(sock->addr.sin_addr),
		g_strerror(error));
  socket_error(sock);

  return;
}

int socket_set_nonblock(socket_ctx_t * sock, int nonblock)
{
  int flags;

  if ((flags = fcntl(sock->sockfd, F_GETFL, 0)) < 0)
    return -1;

  if (nonblock)
    flags |= O_NONBLOCK;
  else
    flags &= ~O_NONBLOCK;

  if (fcntl(sock->sockfd, F_SETFL, flags) < 0)
    return -1;

  return 0;
}

void socket_free(socket_ctx_t * sock)
{
  if (sock->sockfd != -1)
    close(sock->sockfd);

  if (sock->proxy)
    g_free(sock->proxy);

  if (sock->hostname)
    g_free(sock->hostname);

  g_free(sock);
}

void socket_set_data(socket_ctx_t * sock)
{
  int tos = IPTOS_THROUGHPUT;

  g_return_if_fail(sock != NULL);

  if (setsockopt(sock->sockfd, IPPROTO_IP, IP_TOS, &tos, sizeof(tos)) < 0)
    g_error("Couldn't set socket TOS: %s\n", g_strerror(errno));
}

void socket_error(socket_ctx_t * sock)
{
  g_return_if_fail(sock != NULL);

  socket_disconnect(sock);
  sock->callback(NULL, sock->data);
}

void socket_finish(socket_ctx_t * sock)
{
  g_return_if_fail(sock != NULL);

  sock->callback(sock, sock->data);
}

void socket_disconnect(socket_ctx_t * sock)
{
  g_return_if_fail(sock != NULL);

  if (sock->sockfd != -1) {
    close(sock->sockfd);
    sock->sockfd = -1;
  }

  if (sock->watchid != -1) {
    gdk_input_remove(sock->watchid);
    sock->watchid = -1;
  }
}

socket_ctx_t *socket_new()
{
  socket_ctx_t *sock;

  sock = g_new0(socket_ctx_t, 1);
  sock->watchid = -1;
  sock->sockfd = -1;

  return sock;
}
